import cherrypy
from cherrypy.lib import httpauth

try:
    import indigo
except:
    pass

class BadOAuthToken(Exception):
    pass

def check_auth(users=None, encrypt=None, realm=None):
    """If an authorization header contains credentials, return True, else False."""
    # make sure the provided credentials are correctly set
    username = None
    try:
        ah = httpauth.parseAuthorization(cherrypy.request.headers['authorization'])
        if ah["auth_scheme"] != "bearer":
            # We only want to continue if it's not an OAuth token - if it is and oauth is enabled, it will be handled
            # below by the oauth() function
            username = ah["username"]
            if ah is None:
                raise cherrypy.HTTPError(400, 'Bad Request')
            if not encrypt:
                encrypt = httpauth.DIGEST_AUTH_ENCODERS[httpauth.MD5]
            if callable(users):
                try:
                    # backward compatibility
                    users = users() # expect it to return a dictionary
                    # fetch the user password
                    password = users.get(username, None)
                except TypeError as exc:
                    # returns a password (encrypted or clear text)
                    password = users(username)
            else:
                if not isinstance(users, dict):
                    raise ValueError("Authentication users must be a dictionary")
                # fetch the user password
                password = users.get(username, None)
            if password is None:
                cherrypy.request.login = False
                cherrypy.request.failed_login = username
                return False
            # validate the authorization by re-computing it here
            # and compare it with what the user-agent provided
            if httpauth.checkResponse(ah, password, method=cherrypy.request.method,
                                      encrypt=encrypt, realm=realm):
                cherrypy.request.login = username
                return True
    except Exception as exc:
        # TODO: I wonder if we need to log this error? Probably not since it'll just end up with a 401 probably.
        pass
    
    cherrypy.request.login = False
    try:
        cherrypy.request.failed_login = username
    except:
        pass
    return False


def basic_auth(realm, users, encrypt=None):
    """
    If auth fails, raise 401 with a basic authentication header IF digest auth is disabled, otherwise it will
    just return False and let the digest_auth handler have a go at it.

    realm: a string containing the authentication realm.
    users: a dict of the form: {username: password} or a callable returning a dict.
    encrypt: callable used to encrypt the password returned from the user-agent.
             if None it defaults to a md5 encryption.
    """
    if cherrypy.request.login or check_auth(users, encrypt):
        # If we're already logged in or we can authorize then we're good
        return True
    if cherrypy.config.get("tools.digest_auth.on", False):
        # Because basic auth runs before digest, we should just return False if we know digest auth is turned on.
        return False
    else:
        # Otherwise, we are responsible for raising the 401 with the appropriate header so that browsers will
        # prompt for basic auth credentials.
        cherrypy.response.headers['www-authenticate'] = httpauth.basicAuth(realm)
        raise cherrypy.HTTPError(401, "You are not authorized to access that resource")

def digest_auth(realm, users):
    """
    If auth fails, raise 401 with a digest authentication header.

    realm: a string containing the authentication realm.
    users: a dict of the form: {username: password} or a callable returning a dict.
    """
    # If we can validate the creds or something has already validated, then we're good
    if cherrypy.request.login or check_auth(users, realm=realm):
        return True
    else:
        # Because we might be the last shot at validation, we are responsible for throwing the 401 along with the
        # appropriate header so browsers will prompt for digest auth credentials.
        # Note: if digest is disabled then the basic method, which will be called first, will throw the 401 and this
        # method will never get called.
        cherrypy.response.headers['www-authenticate'] = httpauth.digestAuth(realm)
        raise cherrypy.HTTPError(401, "You are not authorized to access that resource")


def oauth():
    """
    Attempt at validating an OAuth protected request. Will return False if it can't validate the token in the
    authorization header.
    """
    return indigo.activePlugin._attempt_oauth()
