####################
# Copyright (c) 2014, Perceptive Automation, LLC. All rights reserved.
# http://www.indigodomo.com
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################

####################
## IMPORTS
import cherrypy
from cherrypy import _cperror
from cherrypy._cpdispatch import RestController

from Cheetah.Template import Template as CheetahTemplate
from Cheetah.Filters import Filter as CheetahFilter

import indigoutil as iu

import os
import threading
import urllib

####################
## CONSTANTS
kTrueStr = u"true"
kFalseStr = u"false"
kEmptyStr = u""

kTextPageStr = u"plain"
kHtmlPageStr = u"html"
kXmlPageStr = u"xml"

_kInternalServerError = u"internal server error"
_kFileNotFoundError = u"undefined page requested"
_kAuthenticationFailed = u"access denied"

####################
kServerStateErrorStr = u"error"	# true if the modal error dialog is showing

_kControlDisabledError = u"remote Web browser access is disabled"
_kControlDisabledDetails = u"To enable remote Web browser access, choose the <em>Indigo->Start Server...</em> menu item from within the client application."

_kNativeAccessDisabledError = u"remote Indigo client access is disabled"
_kNativeAccessDisabledDetails = u"To enable remote Indigo client access, choose the <em>Indigo->Start Server...</em> menu item from within the client application."

_kFeedsDisabledError = u"Indigo feed output support is disabled"
_kFeedsDisabledDetails = u"To enable Indigo feed output, choose the <em>Indigo->Start Server...</em> menu item from within the client application."

_kPageIsListOnlyError = u"page is list-only type -- cannot show graphically"
_kPageIsListOnlyDetails = u"Try adding &amp;asList=1 argument to URL to render this page in list mode."

def null_logger(message):
	pass

####################
class IndigoCheetahFilter(CheetahFilter):
	def filter(self, val, unicodeType=type(kEmptyStr), **kw):
		"""Our template filter does the following with all $placeholder substitutions:

				1) Pass Unicode strings through as raw unicode string. ASCII strings are decoded
				   into unicode so that everything is native unicode objects. We do not want to
				   encode to UTF-8 using this filter (exception: URLs). CherryPy handles that
				   later via the cherrypy encoding filter. Let's keep everything in native
				   unicode format until the last possible moment.

				2) Converts None placeholders into an empty unicode string (u'').

				3) Handles XML/XHTML entity substituion for: < > & ' "

				4) Optionally does URL (and UTF-8) encoding.

			Argument "entityEncoding" values: HTML, XML, URL-node, URL-path, skip (defaults to HTML)

			#1 and #2 above are always processed.
			#3 is processed by default, but can be disabled by passing entityEncoding == skip or URL-*
			#4 is processed if entityEncoding == URL-*
		"""
		if type(val) == unicodeType:
			filtered = val
		elif val is None:
			filtered = kEmptyStr	# u''
		else:	# should not happen, all placeholders should be unicode...
			filtered = str(val).decode("utf-8")
			# Turn on the logs below to see a list of all strings that aren't native
			# unicode. Should have just numbers if everything was created correctly.
			# (i.e. use someVar=u"hello there" instead of someVar="hello there").
			#
			# self.logger.debug("converted to unicode: " + filtered)

		if kw.has_key('entityEncoding'):
			if kw['entityEncoding'] == 'URL-node':
				filtered = filtered.encode("utf-8")			# must utf-8 encode before url escaping
				filtered = urllib.quote(filtered,safe='')	# encode even '/' since val is a node (not path)

				#	-- Workaround for Apache URI slash problem --
				#
				# If our node name has a slash at this point it is encoded like this:
				#
				#	aaa.indigodomo.net/devices/device_with_slash_%2F_here
				#
				# Unfortunately, Apache doesn't like that and returns an immediate 404 error (CherryPy seems
				# fine with it). One option might be to turn on AllowEncodedSlashes in the httpd.conf file:
				#
				#	AllowEncodedSlashes On
				#
				# but the version of Apached we are using doesn't support that yet. So instead, we encode the
				# percentage sign as well, which gives us this:
				#
				#	aaa.indigodomo.net/devices/device_with_slash_%252F_here
				#
				# Both Apache and CherryPy seem to handle this double-encoding okay.
				filtered = filtered.replace("%2F", "%252F")
				return filtered
			elif kw['entityEncoding'] == 'URL-parm':
				filtered = filtered.encode("utf-8")			# must utf-8 encode before url escaping
				filtered = urllib.quote(filtered,safe='')	# encode even '/' since val is a parm (not path)
				return filtered
			elif kw['entityEncoding'] == 'URL-path':
				filtered = filtered.encode("utf-8")			# must utf-8 encode before url escaping
				filtered = urllib.quote(filtered,safe='/')	# skip encoding of '/' since val is a path
				return filtered
			elif kw['entityEncoding'] == 'skip':
				# useful if the placeholder already contains valid HTML or XML that shouldn't be escaped
				return filtered
			elif kw['entityEncoding'] == 'XML':
				return iu.EscapeForXML(filtered)

		# else no entityEncoding defined, default to HTML entity swapping:
		return iu.EscapeForHTML(filtered)

####################
_templateDictLock = threading.Lock()
_templateDict = {}		# cached global template dict; lazy allocate each on first request

class IndigoTemplate(object):
	def __init__(self, templateFilePath):
		self._templateLock = threading.Lock()
		self._template = CheetahTemplate(file = templateFilePath, filter = IndigoCheetahFilter)

	# Jay: Logging update broke this, but it doesn't appear to ever get called...
	# def LogGeneratedCode(self):
	# 	self.logger.info(self._template.generatedModuleCode())
	#
	def AcquireLock(self):
		self._templateLock.acquire()

	def ReleaseLock(self):
		self._templateLock.release()

	def RenderTemplate(self):
		return self._template.respond()

	def __getattr__(self, name):
		return self._template.__getattr__(name)

	def __setattr__(self, name, value):
		if name in ['_templateLock', '_template', '_Log', '_DebugLog']:
			object.__setattr__(self, name, value)
		else:
			self._template.__setattr__(name, value)

	def __delattr__(self, name):
		if name in ['_templateLock', '_template', '_Log', '_DebugLog']:
			object.__delattr__(self, name)
		else:
			self._template.__delattr__(name)

####################
class BaseRequestHandler(RestController):
	""" Handles HTTP page requests. """

	def __init__(self, logger, debug_logger=null_logger):
		self._useRestDispatcher = False
		self.logger = logger
		self._Log = logger
		self._DebugLog = debug_logger

	def _cp_dispatch(self, vpath):
		# Override RestController to not use _cp_dispatch -- IndigoRequestHandler doesn't
		# really want to do REST. Subclasses that want REST should override this.
		if self._useRestDispatcher:
			result = RestController._cp_dispatch(self, vpath)
			if result:		# force parm format to lowercase
				formatKey = cherrypy.config.get('rest_format', 'format')
				formatVal = cherrypy.request.params.get(formatKey, None)
				if formatVal:
					cherrypy.request.params[formatKey] = formatVal.lower()
			return result
		else:
			return None

	def _EnableRestDispatcher(self):
		self._useRestDispatcher = True

	def _NopLog(self, unusedMessage):
		pass

	####################
	def _CheckRestMimeType(self, mimeType):
		if mimeType not in ["xml", "html", "json", "txt"]:
			raise cherrypy.HTTPError(404)

	def _GetMethodArgOverride(self, cur_func, **kw):
		if not kw.has_key('_method'):
			return None			# no method arg override specified
		method_arg = kw['_method'].upper()
		if method_arg == cur_func:
			return None			# method arg override specified, but matches function called (NOP)
		# print "method_arg override " + method_arg
		if method_arg == "GET":
			return self.GET
		elif method_arg == "PUT":
			return self.PUT
		elif method_arg == "POST":
			return self.POST
		elif method_arg == "DELETE":
			return self.DELETE
		elif method_arg == "EXECUTE":
			return self.EXECUTE
		return None

	####################
	def _GetTemplate(self, filePath):
		global _templateDictLock, _templateDict
		_templateDictLock.acquire()
		try:
			curMtime = os.path.getmtime(filePath)
			(cacheMtime, template) = _templateDict.get(filePath, (0, None))
			if template and curMtime > cacheMtime:
				self.logger.debug("-- cached template " + filePath + " stale --")
				template = None
			if not template:
				self.logger.debug("-- building template " + filePath + " --")
				_templateDict[filePath] = (curMtime, IndigoTemplate(filePath))
			return _templateDict[filePath][1]
		finally:
			_templateDictLock.release()

	def _GetAndLockGlobalTemplate(self, templateName):
		filePath = 'templates/' + templateName
		template = self._GetTemplate(filePath)
		template.AcquireLock()
		return template

	def _GetAndLockPluginTemplate(self, templateName):
		USER_STATIC_FOLDER_PATH = cherrypy.config.get("user_static_folder_path")
		filePath = os.path.join(USER_STATIC_FOLDER_PATH, 'plugins/' + templateName)
		template = self._GetTemplate(filePath)
		template.AcquireLock()
		return template

	def _ClearTemplateCache(self):
		global _templateDictLock, _templateDict
		_templateDictLock.acquire()
		try:
			_templateDict = {}
			self.logger.debug("-- cleared template cache")
		finally:
			_templateDictLock.release()

	####################
	def _GetQueryStrings(self, useJS):
		# Calculate version of query string that forces useJS to False.
		queryString = cherrypy.request.query_string
		queryStringStripped = queryString.replace("&useJS="+useJS, "")
		queryStringStripped = queryStringStripped.replace("useJS="+useJS, "")
		if len(queryStringStripped) > 0:
			queryStringStripped = queryStringStripped.rstrip("&") + "&"
		queryStringNoJs = queryStringStripped + "useJS=False"
		queryStringUseJs = queryStringStripped + "useJS=True"
		
		# Also normalize useJS to a boolean.
		if useJS.lower() in [kTrueStr, u"yes", u"1"]:
			useJS = True
		else:
			useJS = False
		return [queryString, queryStringNoJs, queryStringUseJs, useJS]

	####################
	def _FillDefaultTemplateAttrs(self, tmpl):
		tmpl.rootPath = cherrypy.request.script_name
		tmpl.iphone = self._IsiPhone()
		tmpl.android = self._IsAndroid()
		tmpl.mobileWebKit = self._IsMobileWebKit()
		tmpl.hostname = cherrypy.request.base
		tmpl.portnum = str(cherrypy.server.httpPortNum)
		tmpl.requestFromPrism = self._IsRequestFromPrism()
		tmpl.useAlphaImageLoader = self._UseAlphaImageLoader()

	####################
	def _ForceHeadersToNeverCache(self):
		# IE likes to cache XML requests, so we explicitly tell it not to...
		cherrypy.response.headers['Expires'] = 'Mon, 26 Jul 1997 05:00:00 GMT'
		cherrypy.response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, max-age=0'
		cherrypy.response.headers['Pragma'] = 'no-cache'
		# cherrypy.response.headers['Last-Modified'] = gmdate("D, d M Y H:i:s") + "GMT"

	####################
	def _IsRequestFromPrism(self):
		request_host = cherrypy.request.headers.get('Host', '')
		if request_host.find("indigodomo.net") >= 0:
			return True
		# Legacy hosts:
		if request_host.find("goprism.com") >= 0 or request_host.find("prismreflector.com") >= 0:
			return True
		return False

	####################
	def _IsiPhone(self):
		# Detects if the page is being viewed by an iPhone
		if cherrypy.request.headers.has_key('User-Agent'):
			userAgent = cherrypy.request.headers['User-Agent']
			#self.logger.debug("USER AGENT = " + userAgent)
			if userAgent.find("iPhone") >= 0 or userAgent.find("iPod") >= 0 or userAgent.find("iPad") >= 0:
				return True
		return False

	####################
	def _IsAndroid(self):
		# Detects if the page is being viewed by an iPhone
		if cherrypy.request.headers.has_key('User-Agent'):
			userAgent = cherrypy.request.headers['User-Agent']
			#self.logger.debug("USER AGENT = " + userAgent)
			if userAgent.find("Android") >= 0:
				return True
		return False

	####################
	def _IsMobileWebKit(self):
		# Detects if the page is being viewed by webkit on a mobile device (iPhone, Android, etc.)
		if cherrypy.request.headers.has_key('User-Agent'):
			userAgent = cherrypy.request.headers['User-Agent']
			#self.logger.debug("USER AGENT = " + userAgent)
			if userAgent.find("Mobile Safari") >= 0:
				return True
		return False

	####################
	def _UseAlphaImageLoader(self):
		# Windows IE 5.5 and 6.0 cannot display PNG alpha correctly, but we have a work around
		# by writing HTML that tells IE to use a built-in filter. Must only return True here
		# though for IE 5 and 6 on Windows, otherwise browser won't be served the normal <img>
		# tags.
		if cherrypy.request.headers.has_key('User-Agent'):
			userAgent = cherrypy.request.headers['User-Agent']
			#self.logger.debug("USER AGENT = " + userAgent)
			if userAgent.find("Safari") >= 0:
				return False
			if userAgent.find("Gecko") >= 0:
				return False
			if userAgent.find("Opera") >= 0:
				return False
			if userAgent.find("Windows") >= 0:
				if userAgent.find("MSIE 5.") >= 0:
					return True
				if userAgent.find("MSIE 6.") >= 0:
					return True
		return False

	def _UseUniqueUrlToAvoidCache(self):
		# Unbelievably, MSIE 6 will use its cache for 2 seconds regardless of the fact that
		# the HTTP header told it not to cache at all. Adding a unique variable to the URL
		# will prevent this.
		if cherrypy.request.headers.has_key('User-Agent'):
			userAgent = cherrypy.request.headers['User-Agent']
			#self.logger.debug("USER AGENT = " + userAgent)
			if userAgent.find("Windows") >= 0:
				if userAgent.find("MSIE 5.") >= 0:
					return True
				if userAgent.find("MSIE 6.") >= 0:
					return True
		return False

	####################
	def _UpdatePollInterval(self, serverState=None):
		pass	# stub; subclassed by IndigoRequestHandler

	####################
	def _ReturnError(self, pageType, errorMsg, errorDetails = None, logError = True):
		if logError:
			self.logger.info(errorMsg)
		cherrypy.response.headers['Content-Type'] = 'text/' + pageType

		if pageType == kTextPageStr:
			return errorMsg		# just return the single line error (no details)
		else:
			if pageType == kHtmlPageStr:
				tmpl = self._GetAndLockGlobalTemplate('error.html')
			elif pageType == kXmlPageStr:
				tmpl = self._GetAndLockGlobalTemplate('_error_response.xml')
			else:
				return errorMsg		# just return the single line error (no details)
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				if pageType == kXmlPageStr:
					tmpl.serverState = kServerStateErrorStr
					tmpl.pollInterval = self._UpdatePollInterval(kServerStateErrorStr)
					tmpl.openErrorDialog = True
				tmpl.errorMsg = errorMsg
				tmpl.errorDetails = errorDetails
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()

	def _ReturnHtmlError(self, errorMsg, errorDetails = None, logError = True):
		if logError:
			self.logger.info(errorMsg)
		cherrypy.response.headers['Content-Type'] = 'text/' + kHtmlPageStr

		try:
			tmpl = self._GetAndLockGlobalTemplate('error.html')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.errorMsg = errorMsg
				tmpl.errorDetails = errorDetails
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except:
			textErrroMsg = errorMsg + "<br \>\n"
			if errorDetails:
				textErrroMsg = textErrroMsg + errorDetails + "<br \>\n"
			textErrroMsg = textErrroMsg + "(also unable to render HTML error)"
			return textErrroMsg

	####################
	def _ReturnControlDisabled(self, pageType):
		return self._ReturnError(pageType, _kControlDisabledError, _kControlDisabledDetails)

	def _ReturnNativeAccessDisabled(self, pageType):
		return self._ReturnError(pageType, _kNativeAccessDisabledError, _kNativeAccessDisabledDetails)

	def _ReturnFeedsDisabled(self, pageType):
		return self._ReturnError(pageType, _kFeedsDisabledError, _kFeedsDisabledDetails)

	def _ReturnListOnlyError(self, pageType):
		return self._ReturnError(pageType, _kPageIsListOnlyError, _kPageIsListOnlyDetails)

	####################
	def ModErrorPage(self, status, **kwargs):
		if status is None or str(status) is None:
			status_num = 500
		else:
			status_num = int(status.split()[0])
	
		logError = True
		logHeadersAndStack = False		# most useful for internal server errors

		remoteClientIP = getattr(cherrypy.request.remote, 'better_ip', cherrypy.request.remote.ip)
		if status_num == 404:
			errStr = _kFileNotFoundError
		elif status_num == 401:
			errStr = _kAuthenticationFailed
			if getattr(cherrypy.request, "failed_login", False):
				remoteClientIP = cherrypy.request.failed_login + " @ " + remoteClientIP
			else:	# if no failed_login, then this is probably the initial page request so do not log
				logError = False
		elif status_num == 500:
			errStr = _kInternalServerError
			logHeadersAndStack = True
		else:
			errStr = unicode(str(status), "utf-8")
	
		# Log basic error to Indigo (Event Log window) and also to CherryPy.
		if logError:
			self.logger.info(errStr + " \"" + cherrypy.url() + "\" from " + remoteClientIP)
	
		# For internal server errors, also log the request headers and stack trace.
		if logHeadersAndStack:
			h = ["  %s: %s" % (k, v) for k, v in cherrypy.request.header_list]
			headersAndStackTrace = "\n\n"
			headersAndStackTrace += 'Request Headers:\n' + '\n'.join(h)
			headersAndStackTrace += '\n\n'
			headersAndStackTrace += _cperror.format_exc()
			self.logger.info(headersAndStackTrace)
	
		responseHtml = self._ReturnHtmlError(errStr, None, False)
		return responseHtml
