####################
# Copyright (c) 2021, Perceptive Automation, LLC. All rights reserved.
# http://www.indigodomo.com
#
# Redistribution of this source file, its binary forms, and images are not allowed.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################

####################
## IMPORTS
import socket
import xml.dom.minidom

import base64
try:
	from hashlib import sha1 as sha		# for python 2.5+
except ImportError:
	from sha import new as sha			# but support older versions as well

import os
import threading

####################
## GLOBAL CONSTANTS
kConn_ClientWebServer = u'1'
kConn_ClientFlash = u'2'
kConn_ClientSQL = u'3'

## LOCAL CONSTANTS
_kTrueStr = u"true"
_kFalseStr = u"false"
_kOnStr = u"on"
_kOffStr = u"off"

# The native XML server. You could potentially run the HTTP server on a different
# (not localhost) machine than the XML server. However, some additional handshaking
# and authentication will need to be added.
_kXmlServerAddress = '127.0.0.1'

# Client And Server packet types:
_kAuthenticateType = u"Authenticate"

# Client packet types:
_kCommandType = u"Command"
_kRequestType = u"Request"

# Server packet types:
_kBroadcastType = u"Broadcast"
_kResponseType = u"Response"

# Connection states:
_kConnGood = u"good"
_kConnPending = u"pending"
_kConnClosed = u"connection with Indigo Server closed -- try restarting Indigo Server"
_kConnFailed = u"failed to connect with Indigo Server -- try restarting Indigo Server"
_kConnAuthenDataBad = u"failed to retrieve authentication data from Indigo Server -- try restarting Indigo Server"
_kConnAuthenFailed = u"failed to authenticate with Indigo Server -- "
_kConnLicenseMaxed = u'your current license of Indigo Server does not allow remote web access. to enable web access, <a href="http://www.indigodomo.com/purchase-upgrade" target="_blank">upgrade</a> your license'

####################
## GLOBALS
_gXmlDoc = xml.dom.minidom.getDOMImplementation().createDocument(None, None, None)

####################
## EXCEPTIONS
class ConnError(Exception):
	def __init__(self, value):
		self.value = value	# unicode string

class InternalError(Exception):
	def __init__(self, value):
		self.value = value	# unicode string

class ServerError(Exception):
	def __init__(self, value):
		self.value = value	# unicode string

####################
class UserPerm:
	""" User permissions object, including HTTP digest password hash """
	def __init__(self, webHash, permFlags):
		self.webHash = webHash
		self.permFlags = permFlags

####################
def GetChildElementByTagName(xmlObj, childName):
	""" Like getElementsByTagName(), but not recursive """
	if not xmlObj or not xmlObj.childNodes:
		return None

	for childNode in xmlObj.childNodes:
		if childNode.nodeType == childNode.ELEMENT_NODE and childNode.localName == childName:
			return childNode
	return None

####################
class XmlNode:
	""" Wrapper around minidom object for easy access to Indigo structured XML. """

	def __init__(self, xmlObj=None):
		if xmlObj:
			self.xmlObj = xmlObj
		else:
			# must be a new data node -- use a Data dict as default
			self.xmlObj = _gXmlDoc.createElement(u"Data")
			self.xmlObj.setAttribute(u"type", u"dict")

	def AddStringNode(self, nodeName, nodeVal):
		if not isinstance(nodeName, unicode):
			nodeName = unicode(nodeName, "utf-8")
		if not isinstance(nodeVal, unicode):
			nodeVal = unicode(nodeVal, "utf-8")

		newNode = _gXmlDoc.createElement(nodeName)
		newNode.setAttribute(u"type", u"string")
		textNode = _gXmlDoc.createTextNode(nodeVal)
		newNode.appendChild(textNode)
		self.xmlObj.appendChild(newNode)

	def AddDictNode(self, nodeName):
		newNode = _gXmlDoc.createElement(nodeName)
		newNode.setAttribute(u"type", u"dict")
		self.xmlObj.appendChild(newNode)
		return XmlNode(newNode)

	def AddVectorNode(self, nodeName):
		newNode = _gXmlDoc.createElement(nodeName)
		newNode.setAttribute(u"type", u"vector")
		self.xmlObj.appendChild(newNode)
		return XmlNode(newNode)

	def CleanUp(self):
		self.xmlObj.unlink()
		self.xmlObj = None

	def GenerateXml(self):	# always returns utf-8 encoded data
		return self.xmlObj.toxml("utf-8")

	####################
	def GetElemStr(self):
		# returns our string value
		#
		return self.xmlObj.childNodes[0].data

	def GetElemBool(self):
		# returns our boolean value
		#
		elemStr = self.GetElemStr()
		if elemStr and elemStr == _kTrueStr:
			return True
		return False

	def GetElemInt(self):
		# returns our integer value
		#
		elemStr = self.GetElemStr()
		if elemStr:
			return int(elemStr)
		return 0

	####################
	def GetChildElemStr(self, childName):
		# returns string value of child node elem
		#
		childElem = GetChildElementByTagName(self.xmlObj, childName)
		if childElem and childElem.childNodes:
			return childElem.childNodes[0].data
		return None

	def GetChildElemBool(self, childName, defaultVal = False):
		# returns boolean value of child node elem
		#
		childElem = self.GetChildElemStr(childName)
		if childElem:
			if childElem in (_kTrueStr, _kOnStr):
				return True
			else:
				return False
		return defaultVal

	def GetChildElemInt(self, childName, defaultVal = 0):
		# returns integer value of child node elem
		#
		childElem = self.GetChildElemStr(childName)
		if childElem:
			return int(childElem)
		return defaultVal

	def GetChildElemIntPair(self, childName):
		# returns pair of integer values
		#
		childElem = self.GetChildElemStr(childName)
		if childElem:
			(n1, n2) = str(childElem).split(" ")
			return [int(n1), int(n2)]
		return [0, 0]

	def GetChildElemIntQuad(self, childName):
		# returns quad of integer values
		#
		childElem = self.GetChildElemStr(childName)
		if childElem:
			(n1, n2, n3, n4) = str(childElem).split(" ")
			return [int(n1), int(n2), int(n3), int(n4)]
		return [0, 0, 0, 0]

	def GetChildElemHtmlColor(self, childName, defaultColor):
		# returns HTML color string of child node elem
		#
		childElem = self.GetChildElemStr(childName)
		if childElem:
			return u"#" + childElem.replace(u" ", u"")
		return defaultColor

	def GetChildElemNode(self, childName):
		# returns child node (XmlNode) elem
		#
		childElem = GetChildElementByTagName(self.xmlObj, childName)
		return XmlNode(childElem)

	####################
	def GetListOfChildElems(self, childName):
		# returns list of child node (XmlNode) elems
		#
		xmlVector = []
		xmlObjList = self.xmlObj.getElementsByTagName(childName)
		for xmlElem in xmlObjList:
			xmlVector.append(XmlNode(xmlElem))
		return xmlVector

####################
class Packet:
	""" Communication Packet used to send and receive commands, requests, and replies. """

	def __init__(self, packetTypeStr, packetNameStr, packetDataVal=None, packetDataNode=None):
		self.typeStr = packetTypeStr
		self.nameStr = packetNameStr

		# packetDataVal is an optimized way to create a packet with a simple data
		# string value. Use this for most outgoing command packets. Incoming packets
		# will never use this.
		if packetDataVal and packetDataNode:
			raise InternalError, u"can only use either data node or data value, not both"

		self.xmlValue = None
		self.xmlNode = None
		if packetDataVal:
			if not isinstance(packetDataVal, unicode):
				self.xmlValue = unicode(packetDataVal, "utf-8")
			else:
				self.xmlValue = packetDataVal
			# handle XML entity substitutions:
			self.xmlValue = self.xmlValue.replace("&", "&amp;")
			self.xmlValue = self.xmlValue.replace("<", "&lt;")
			self.xmlValue = self.xmlValue.replace(">", "&gt;")
		elif packetDataNode:
			self.xmlNode = packetDataNode

	def IsBroadcastType(self):
		if self.typeStr == _kBroadcastType:
			return True
		return False

	def IsLogStreamBroadcast(self):
		if self.IsBroadcastType() and self.nameStr == u"LogStream":
			return True
		return False

	def CleanUp(self):
		self.typeStr = None
		self.nameStr = None
		if self.xmlNode:
			self.xmlNode.CleanUp()
			self.xmlNode = None

	def GenerateXml(self):	# always returns utf-8 encoded data
		xmlStream = '<Packet type="dict"><Type type="string">'
		xmlStream += self.typeStr.encode("utf-8")
		xmlStream += '</Type><Name type="string">'
		xmlStream += self.nameStr.encode("utf-8")
		xmlStream += '</Name>'
		if self.xmlValue:	# simple case -- data value is a string
			xmlStream += '<Data type="string">'
			xmlStream += self.xmlValue.encode("utf-8")
			xmlStream += '</Data>'
		elif self.xmlNode:	# complex case -- data node is an XmlNode
			xmlStream += self.xmlNode.GenerateXml()
		xmlStream += '</Packet>'
		xmlStream += '\0'
		return xmlStream

####################
class Connection:
	""" Document me. """

	####################
	def __init__(self, connType, connName, connCompleteFunc, portNum, logFunc, debugLogFunc):
		self._portNum = portNum
		self._readBuf = ""	# incoming socket data
		self._readBufSize = 8192
		self._connType = connType
		self._connName = connName
		self._connCompleteFunc = connCompleteFunc
		self._connState = _kConnPending
		self._connLock = threading.RLock()

		self._serverDbName = u"unknown"
		self._serverLicenseLevel = 0
		self._serverVersion = u"unknown"

		self._Log = logFunc
		if debugLogFunc:
			self._DebugLog = debugLogFunc
		else:
			self._DebugLog = self._NopLog

	def _NopLog(self, unusedMessage):
		pass

	####################
	def _RaiseConnError(self, newConnState):
		self._connState = newConnState
		raise ConnError, self._connState

	def _RaiseServerError(self, errorMsg):
		raise ServerError, errorMsg

	####################
	def _Connect(self):
		# Create an Indigo Server connection.
		#
		# We have the config setup to not use threading, so this should
		# always create only a single connection with the Indigo Server.
		# If we ever use threading, then we need to create multiple per
		# thread connections or mutex lock around calls to the server.
		self._Log("connecting to Indigo server")

		try:
			self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self._socket.connect((_kXmlServerAddress, self._portNum))
			self._socket.setblocking(1)
		except Exception as exc:
			self._RaiseConnError(_kConnFailed)

		userID = str(os.getuid())
		salt = "fleasbeescheesekneesteaselouise"				# some useless salt
		userWithSalt = userID + salt
		hashDigest = sha(userWithSalt).digest()
		hashB64encoded = base64.encodestring(hashDigest)[:-1]	# drop LF added at end of string (grrr)

		doChangeBroadcasts = False
		if self._connType == kConn_ClientSQL:
			doChangeBroadcasts = True

		dataNode = XmlNode()
		dataNode.AddStringNode(u"ClientName", self._connName)
		dataNode.AddStringNode(u"ClientType", self._connType)
		dataNode.AddStringNode(u"ProcessUserID", hashB64encoded)
		self._SendPacket(Packet(_kAuthenticateType, u"KnockKnock", packetDataNode=dataNode))

		responsePacket = self._GetNextPacket(_kAuthenticateType)
		if responsePacket.nameStr == u"Success":
			xmlNode = responsePacket.xmlNode
			self._serverDbName = xmlNode.GetChildElemStr(u"DatabaseName")
			self._serverLicenseLevel = xmlNode.GetChildElemInt(u"LicenseLevel")
			self._serverVersion = xmlNode.GetChildElemStr(u"ServerVersionString")

			responsePacket.CleanUp()
			self._connState = _kConnGood
			self._connCompleteFunc()

			if doChangeBroadcasts:
				doBroadcastsNode = XmlNode()
				subscribeListsNode = doBroadcastsNode.AddVectorNode(u"Data")
				subscribeListsNode.AddStringNode(u"Name", "AddedDevice")
				subscribeListsNode.AddStringNode(u"Name", "RemovedDevice")
				subscribeListsNode.AddStringNode(u"Name", "ReplacedDevice")
				subscribeListsNode.AddStringNode(u"Name", "ReplacedAllDevices")
				subscribeListsNode.AddStringNode(u"Name", "AddedVariable")
				subscribeListsNode.AddStringNode(u"Name", "RemovedVariable")
				subscribeListsNode.AddStringNode(u"Name", "ReplacedVariable")
				subscribeListsNode.AddStringNode(u"Name", "LogStream")
				self.SendCommandPacket(u"SubscribeToServerBroadcasts", dataNode=subscribeListsNode)
		elif responsePacket.nameStr == u"LicenseMaxed":
			responsePacket.CleanUp()
			self._RaiseConnError(_kConnLicenseMaxed)
		else:
			errStr = _kConnAuthenFailed + responsePacket.nameStr
			responsePacket.CleanUp()
			self._RaiseConnError(errStr)

	def _Disconnect(self):
		# Disconnect from the Indigo Server.
		self._Log("disconnecting from Indigo server")
		try:
			self._socket.close()
		except:
			pass	# ignore any socket close failures
		self._connState = _kConnClosed

	####################
	def _GetNextPacket(self, waitForPacketType=None, waitForPacketName=None):
		while True:
			packetBuf = self._ReadPacketBuffer()
# Useful for debugging what is received from the server:
#			print
#			print packetBuf
#			print
			packetDom = xml.dom.minidom.parseString(packetBuf)
			packetElem = GetChildElementByTagName(packetDom, u"Packet")

			packetType = GetChildElementByTagName(packetElem, u"Type").childNodes[0].data
			packetName = GetChildElementByTagName(packetElem, u"Name").childNodes[0].data
			if waitForPacketType == None or waitForPacketType == packetType:
				if waitForPacketName == None or waitForPacketName == packetName:
					packetData = GetChildElementByTagName(packetElem, u"Data")
					if packetData:
						return Packet(packetType, packetName, packetDataNode=XmlNode(packetData))

					packetError = GetChildElementByTagName(packetElem, u"Error")
					if packetError:
						errorNode = XmlNode(packetError)
						errStr = errorNode.GetChildElemStr(u"Description")
						if not errStr:
							errStr = errorNode.GetChildElemStr(u"Type")
						errorNode.CleanUp()
						self._RaiseServerError(errStr)
					self._RaiseServerError(u"bad response from server")
				else:
					self._DebugLog("ignoring packet " + packetType + ":" + packetName + " (name mismatch)")
			else:
				self._DebugLog("ignoring packet " + packetType + ":" + packetName + " (type mismatch)")

	def _ReadPacketBuffer(self):
		# Get the next raw XML packet buffer from socket
		tempBuf = self._readBuf

		# Packet end is defined by either \n or NULL byte
		nextLine = tempBuf.find('\n')
		nextNull = tempBuf.find('\0') 
		if nextLine >= 0 or nextNull >= 0:
			if nextLine >= 0 and nextNull >=0:
				if nextLine < nextNull:
					nextPacket = nextLine
				else:
					nextPacket = nextNull
			elif nextLine >= 0:
				nextPacket = nextLine
			else:	# nextNull >= 0:
				nextPacket = nextNull
			nextPacket += 1
			self._readBuf = tempBuf[nextPacket:]
			return tempBuf[:nextPacket-1]
		bufList = []
		if tempBuf:
			bufList.append(tempBuf)
		self._readBuf = ""
		while True:
			try:
				tempBuf = self._socket.recv(self._readBufSize)
			except:
				self._RaiseConnError(_kConnClosed)
			if not tempBuf:
				self._RaiseConnError(_kConnClosed)

			bufList.append(tempBuf)
			nextLine = tempBuf.find('\n')
			nextNull = tempBuf.find('\0')
			if nextLine >= 0 or nextNull >= 0:
				if nextLine >= 0 and nextNull >=0:
					if nextLine < nextNull:
						nextPacket = nextLine
					else:
						nextPacket = nextNull
				elif nextLine >= 0:
					nextPacket = nextLine
				else:	# nextNull >= 0:
					nextPacket = nextNull
				nextPacket += 1
				self._readBuf = tempBuf[nextPacket:]
				bufList[-1] = tempBuf[:nextPacket-1]
				break
		return "".join(bufList)

	def _SendPacket(self, packet):
		xmlStream = packet.GenerateXml()
		xmlStream = xmlStream.replace("\t", "&#x09;")
		xmlStream = xmlStream.replace("\r", "&#x0D;")
		xmlStream = xmlStream.replace("\n", "&#x0A;")
# Useful for debugging what is sent to the server:
#		print
#		print xmlStream
#		print
		try:
			self._socket.sendall(xmlStream)
		except:
			self._RaiseConnError(_kConnClosed)

	####################
	def SendCommandPacket(self, commandName, dataVal=None, dataNode=None):
		self.CheckConnection()
		self._connLock.acquire()
		try:
			self._SendPacket(Packet(_kCommandType, commandName, packetDataVal=dataVal, packetDataNode=dataNode))
		finally:
			self._connLock.release()

	def SendRequestPacket(self, requestName, dataVal=None, dataNode=None):
		self.CheckConnection()
		self._connLock.acquire()
		try:
			self._SendPacket(Packet(_kRequestType, requestName, packetDataVal=dataVal, packetDataNode=dataNode))
			responsePacket = self._GetNextPacket(_kResponseType, requestName)
		finally:
			self._connLock.release()
		return responsePacket

	####################
	def SendCommandRaw(self, rawXmlData):
		self._connLock.acquire()
		try:
			try:
				if not isinstance(rawXmlData, unicode):
					rawXmlData = unicode(rawXmlData, "utf-8")
				xmlStream = rawXmlData.encode("utf-8")
				xmlStream += '\0'
				self._socket.sendall(xmlStream)
			except:
				self._RaiseConnError(_kConnClosed)
		finally:
			self._connLock.release()

	def SendRequestRaw(self, rawXmlData):
		self._connLock.acquire()
		try:
			try:
				if not isinstance(rawXmlData, unicode):
					rawXmlData = unicode(rawXmlData, "utf-8")
				xmlStream = rawXmlData.encode("utf-8")
				xmlStream += '\0'
				self._socket.sendall(xmlStream)
				return self._ReadPacketBuffer()
			except:
				self._RaiseConnError(_kConnClosed)
		finally:
			self._connLock.release()

	####################
	def CheckConnection(self):
		# Will throw if connection is bad AND a reconnection attempt fails.
		#
		if self._connState != _kConnGood:
			self._connLock.acquire()
			try:
				self._connState = _kConnPending
				self._Connect()		# throws ConnError if there is a failure
			finally:
				self._connLock.release()

	def IsConnectionGood(self):
		return self._connState == _kConnGood

	def GetConnectionState(self):
		return self._connState

	def Disconnect(self):
		self._connLock.acquire()
		try:
			self._Disconnect()
		finally:
			self._connLock.release()

	####################
	def ServerWriteLog(self, logMessage):
		if not self.IsConnectionGood():
			return		# silently return -- do not try to establish a connection just to log
		dataNode = XmlNode()
		dataNode.AddStringNode(u"TypeStr", self._connName)
		dataNode.AddStringNode(u"Message", logMessage)
		self.SendCommandPacket(u"WriteLog", dataNode=dataNode)

