####################
# Copyright (c) 2014, Perceptive Automation, LLC. All rights reserved.
# http://www.indigodomo.com
#
# Redistribution of this source file, its binary forms, and images are not allowed.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################

####################
## IMPORTS
try:
    import indigo
except:
    pass

import indigoconn as ic
import indigoutil as iu

import calendar, time
from types import TupleType, DictType, ListType

import os

####################
## CONSTANTS
_kTrueStr = u"true"
_kFalseStr = u"false"
_kEmptyStr = u""

# Indigo Epoch is 01/01/2000 UTC. Const below is handy for converting to/from UNIX Epoch.
_kEpochOffset = calendar.timegm((2000, 1, 1, 0, 0, 0))

# Space between caption and its state image; must match internal value used by Indigo page editor
_kCaptionBufferSpace = 3

# Mirrored from CPageElem.h:
_kGetPageFlags_None									= 0x00000000
#		----------------
_kGetPageFlags_IgnorePageElems						= 0x00000001
_kGetPageFlags_IgnoreActions						= 0x00000002
#		----------------
_kGetPageFlags_RetrieveAsList						= 0x00000100
#		----------------
_kGetPageFlags_EmbedTargetValues					= 0x00010000
_kGetPageFlags_EmbedTargetElems						= 0x00020000
_kGetPageFlags_EmbedTargetElemsForPopupActionOnly	= 0x00040000

####################
## GLOBALS

####################
## EXCEPTIONS
class ControlDisabled(Exception):
	def __init__(self, value=None):
		self.value = value	# unicode string

####################
def SortByAttr(listElem, attr):
	tempList = [ (getattr(elem, attr).lower(), index, elem) for index, elem in enumerate(listElem) ]
	tempList.sort()
	return [ elem[-1] for elem in tempList ]

####################
def GetAddressStrForINSTEON(addressVal):
	addressStr = u"%02X" % ((addressVal >> 16) & 0xFF)
	addressStr += u".%02X" % ((addressVal >> 8) & 0xFF)
	addressStr += u".%02X" % (addressVal & 0xFF)
	return addressStr

def GetAddressStrForX10(addressVal):
	houseCodes = (u"M",u"E",u"C",u"K",u"O",u"G",u"A",u"I",u"N",u"F",u"D",u"L",u"P",u"H",u"B",u"J")
	deviceCodes = (u"13",u"5",u"3",u"11",u"15",u"7",u"1",u"9",u"14",u"6",u"4",u"12",u"16",u"8",u"2",u"10")
	addressStr = houseCodes[(addressVal >> 4) & 0x0F]
	addressStr += deviceCodes[addressVal & 0x0F]
	return addressStr

####################
class LeafAttribute:
	def __init__(self, key, value, editable, skipValueEncoding = False):
		self.key = key
		self.value = value
		self.editable = editable
		self.skipValueEncoding = skipValueEncoding

class DictElem:
	""" Document me. """
	def __init__(self, typeName):
		self._typeName = typeName

	def _GenerateRawXml(self):
		# A bit of a hack to allow a nested list elem inside a dict. Only works for read-only
		# links, and always generates XML for the nested element regardless of RESTful file
		# extension requested. Currently used for INSTEON link records.
		rawXml = u"\t\t<" + self._typeName + ">\n"

		keys = self._GetPublicKeys()
		for key in keys:
			value = getattr(self, key, None)
			if value is not None and not callable(value):
				if type(value) == type(_kEmptyStr):
					valueUni = value			# already unicode
				else:
					valueUni = str(value).decode("utf-8")

				rawXml += u"\t\t\t<" + key + ">"
				rawXml += iu.EscapeForXML(valueUni)
				rawXml += u"</" + key + ">\n"

		rawXml += "\t\t</" + self._typeName + ">\n"
		return rawXml

	def _GetPublicKeys(self):
		keys = self.__dict__.keys()
		keys.sort()

		# Remove all underscore attribute keys and move 'name' to top of list
		keys2 = []
		for key in keys:
			if key == "name":
				keys2.insert(0, key)
			elif not key.startswith("_"):
				keys2.append(key)
		return keys2

	def GetTypeName(self):
		return self._typeName

	def IsEditable(self):
		editableAttrs = getattr(self, '_editableAttrs', [])
		return len(editableAttrs) > 0

	def StreamPublicKeyVals(self):
		keyValList = []
		keys = self._GetPublicKeys()
		editableAttrs = getattr(self, '_editableAttrs', [])
		for key in keys:
			value = getattr(self, key, None)
			if value is None:
				pass
			elif type(value) == ListType:
				# A bit of a hack to allow a nested list elem inside a dict. Only works for read-only
				# links, and always generates XML for the nested element regardless of RESTful file
				# extension requested. Currently used for INSTEON link records.
				valueUni = u"\n"
				for child in value:
					valueUni = valueUni + child._GenerateRawXml()
				valueUni = valueUni + u"\t"

				leafAttr = LeafAttribute(key, valueUni, False, True)
				keyValList.append(leafAttr)
			elif not callable(value):
				if type(value) == type(_kEmptyStr):
					valueUni = value			# already unicode
				else:
					valueUni = str(value).decode("utf-8")

				editable = key in editableAttrs

				leafAttr = LeafAttribute(key, valueUni, editable)
				keyValList.append(leafAttr)
		return keyValList

	def __iter__(self):
		devDict = dict()
		for leafAttr in self.StreamPublicKeyVals():
			if leafAttr.value == "True":
				yield(leafAttr.key, True)
			elif leafAttr.value == "False":
				yield(leafAttr.key, False)
			elif iu.isIntCompat(leafAttr.value):
				yield(leafAttr.key, int(leafAttr.value))
			elif iu.isFloatCompat(leafAttr.value):
				yield(leafAttr.key, float(leafAttr.value))
			else:
				yield(leafAttr.key, leafAttr.value)

####################
class ActionGroup(DictElem):
	""" Document me. """

	def __init__(self, id, name, desc, displayInUI):
		DictElem.__init__(self, u'action')
		self.id = id
		self.name = name
		self.desc = desc
		self.displayInUI = displayInUI

	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def ActionGroupFromXml(xmlGroup):
	id = xmlGroup.GetChildElemInt(u"ID", 0)
	name = xmlGroup.GetChildElemStr(u"Name")
	desc = xmlGroup.GetChildElemStr(u"Description")
	displayInUI = xmlGroup.GetChildElemBool(u"DisplayInRemoteUI", True)		# may be None; defaults to True

	return ActionGroup(id, name, desc, displayInUI)

####################
#######
def DeviceThermoCurModeToStr(mode):
	mode = int(mode)
	if mode == 0:
		return u"all off"
	elif mode == 1:
		return u"heat on"
	elif mode == 2:
		return u"cool on"
	elif mode == 3:
		return u"auto on"
	elif mode == 4:
		return u"program heat on"
	elif mode == 5:
		return u"program cool on"
	elif mode == 6:
		return u"program auto on"
	return u"unknown"

def DeviceThermoStrToCurMode(modeStr):
	modeStr = modeStr.lower()
	if modeStr == "all off":
		return 0
	elif modeStr == "heat on":
		return 1
	elif modeStr == "cool on":
		return 2
	elif modeStr == "auto on":
		return 3
	elif modeStr == "program heat on":
		return 4
	elif modeStr == "program cool on":
		return 5
	elif modeStr == "program auto on":
		return 6
	return -1

def DeviceThermoFanModeToStr(mode):
	mode = int(mode)
	if mode == 0:
		return u"auto on"
	elif mode == 1:
		return u"always on"
	return u"unknown"

def DeviceThermoStrToFanMode(modeStr):
	modeStr = modeStr.lower()
	if modeStr == "auto on":
		return 0
	elif modeStr == "always on":
		return 1
	return -1

#######
class Device(DictElem):
	""" Document me. """

	####################
	def __init__(self, id, name, desc, classID, devProtocol, type, typeFlags, versByte, address, lastChanged, displayInUI, displayRawState, displayLongState, hasStateToDisplay, xmlDev):
		DictElem.__init__(self, u'device')
		self.id = id
		self.name = name
		self.desc = desc
		self.classID = classID
		self.devProtocol = devProtocol
		self.address = address
		self.type = type
		self.typeFlags = typeFlags
		self.versByte = versByte
		self.displayRawState = displayRawState
		self.displayLongState = displayLongState

		self.folderID = xmlDev.GetChildElemInt(u"FolderID")

		####################
		if devProtocol == 0:		# X10
			self.addressStr = GetAddressStrForX10(self.address)
		elif devProtocol == 1:		# INSTEON
			self.addressStr = GetAddressStrForINSTEON(self.address)
		else:
			self.addressStr = u"unknown"

		####################
		self.typeIsSprinkler = False
		self.typeSupportsSprinkler = False		# legacy for RESTful API, remove eventually
		self.activeZone = None
		self.zoneNames = None
		if self.classID == 5:
			self.typeIsSprinkler = True
			self.typeSupportsSprinkler = True	# legacy for RESTful API, remove eventually
			self.activeZone = xmlDev.GetChildElemInt(u"ActiveSprinkler")
			propertiesDict = xmlDev.GetChildElemNode(u"Properties")
			if propertiesDict:
				self.zoneNames = propertiesDict.GetChildElemStr(u"ZoneNames")
				self.previousZoneDurations = propertiesDict.GetChildElemStr(u"PreviousZoneDurations")
				self.previousScheduleZoneIndex = propertiesDict.GetChildElemInt(u"PreviousScheduleZoneIndex")
				self.scheduledZoneDurations = propertiesDict.GetChildElemStr(u"ScheduledZoneDurations")
				self.currentScheduleZoneIndex = propertiesDict.GetChildElemInt(u"CurrentScheduleZoneIndex")
				self.pauseScheduleZoneIndex = propertiesDict.GetChildElemInt(u"PauseScheduleZoneIndex")
			else:
				self.zoneNames = _kEmptyStr
				self.previousZoneDurations = _kEmptyStr
				self.previousScheduleZoneIndex = 0
				self.scheduledZoneDurations = _kEmptyStr
				self.currentScheduleZoneIndex = 0
				self.pauseScheduleZoneIndex = 0

		####################
		self.typeIsMultiIO = False
		self.typeSupportsIO = False			# legacy for RESTful API, remove eventually
		self.inputBinaryStates = None
		self.inputAnalogStates = None
		self.inputSensorStates = None
		self.outputBinaryStates = None
		self.outputTimers = None
		self.labelInputBinaryOff = None
		self.labelInputBinaryOn = None
		self.labelOutputBinaryOff = None
		self.labelOutputBinaryOn = None
		if self.classID == 4:
			self.typeIsMultiIO = True
			self.typeSupportsIO = True		# legacy for RESTful API, remove eventually
			propertiesDict = xmlDev.GetChildElemNode(u"Properties")
			if propertiesDict:
				self.inputBinaryStates = propertiesDict.GetChildElemStr(u"InputBinaryStates")
				self.inputAnalogStates = propertiesDict.GetChildElemStr(u"InputAnalogValues")
				self.inputSensorStates = propertiesDict.GetChildElemStr(u"InputSensorValues")
				self.outputBinaryStates = propertiesDict.GetChildElemStr(u"OutputBinaryStates")
				self.outputTimers = propertiesDict.GetChildElemStr(u"OutputTimers")
				self.labelInputBinaryOff = propertiesDict.GetChildElemStr(u"LabelInputBinaryOff")		# can be none
				self.labelInputBinaryOn = propertiesDict.GetChildElemStr(u"LabelInputBinaryOn")			# can be none
				self.labelOutputBinaryOff = propertiesDict.GetChildElemStr(u"LabelOutputBinaryOff")		# can be none
				self.labelOutputBinaryOn = propertiesDict.GetChildElemStr(u"LabelOutputBinaryOn")		# can be none

		####################
		# Thermostat states
		self.typeIsHVAC = False
		self.typeSupportsHVAC = False			# legacy for RESTful API, remove eventually
		self.setpointCool = None
		self.setpointHeat = None
		self.inputTemperatureVals = None
		self.inputHumidityVals = None
		self.hvacCurrentMode = None
		self.hvacFanMode = None
		self.hvacCoolerIsOn = None
		self.hvacDehumidifierIsOn = None
		self.hvacFanIsOn = None
		self.hvacFanModeIsAlwaysOn = None
		self.hvacFanModeIsAuto = None
		self.hvacHeaterIsOn = None
		self.hvacHumidifierIsOn = None
		self.hvacOperationModeIsAuto = None
		self.hvacOperationModeIsCool = None
		self.hvacOperationModeIsHeat = None
		self.hvacOperationModeIsOff = None
		self.hvacOperationModeIsProgramAuto = None
		self.hvacOperationModeIsProgramCool = None
		self.hvacOperationModeIsProgramHeat = None
		self.displayHumidityInRemoteUI = None
		if self.classID == 6:
			self.typeIsHVAC = True
			self.typeSupportsHVAC = True		# legacy for RESTful API, remove eventually
			self.setpointCool = xmlDev.GetChildElemStr(u"ActiveSetpointCool")
			self.setpointHeat = xmlDev.GetChildElemStr(u"ActiveSetpointHeat")
			propertiesDict = xmlDev.GetChildElemNode(u"Properties")
			if propertiesDict:
				self.inputTemperatureVals = propertiesDict.GetChildElemStr(u"InputTemperatureValues")
				self.inputHumidityVals = propertiesDict.GetChildElemStr(u"InputHumidityValues")

				currentMode = propertiesDict.GetChildElemInt(u"CurrentOperationMode")
				self.hvacCurrentMode = DeviceThermoCurModeToStr(currentMode)

				fanMode = propertiesDict.GetChildElemInt(u"CurrentFanMode")
				self.hvacFanMode = DeviceThermoFanModeToStr(fanMode)

				self.displayHumidityInRemoteUI = propertiesDict.GetChildElemBool(u"DisplayHumidityInRemoteUI")

				statesDict = xmlDev.GetChildElemNode(u"States")
				if statesDict:
					self.hvacCoolerIsOn = statesDict.GetChildElemBool(u"hvacCoolerIsOn")
					self.hvacDehumidifierIsOn = statesDict.GetChildElemBool(u"hvacDehumidifierIsOn")
					self.hvacFanIsOn = statesDict.GetChildElemBool(u"hvacFanIsOn")
					self.hvacFanModeIsAlwaysOn = statesDict.GetChildElemBool(u"hvacFanModeIsAlwaysOn")
					self.hvacFanModeIsAuto = statesDict.GetChildElemBool(u"hvacFanModeIsAuto")
					self.hvacHeaterIsOn = statesDict.GetChildElemBool(u"hvacHeaterIsOn")
					self.hvacHumidifierIsOn = statesDict.GetChildElemBool(u"hvacHumidifierIsOn")
					self.hvacOperationModeIsAuto = statesDict.GetChildElemBool(u"hvacOperationModeIsAuto")
					self.hvacOperationModeIsCool = statesDict.GetChildElemBool(u"hvacOperationModeIsCool")
					self.hvacOperationModeIsHeat = statesDict.GetChildElemBool(u"hvacOperationModeIsHeat")
					self.hvacOperationModeIsOff = statesDict.GetChildElemBool(u"hvacOperationModeIsOff")
					self.hvacOperationModeIsProgramAuto = statesDict.GetChildElemBool(u"hvacOperationModeIsProgramAuto")
					self.hvacOperationModeIsProgramCool = statesDict.GetChildElemBool(u"hvacOperationModeIsProgramCool")
					self.hvacOperationModeIsProgramHeat = statesDict.GetChildElemBool(u"hvacOperationModeIsProgramHeat")

		####################
		self.typeSupportsOnOff = False
		self.typeSupportsSensorValue = False
		self.typeSupportsEnergyMeter = False
		self.typeIsSpeedControl = False
		self.typeSupportsSpeedControl = False			# legacy for RESTful API, remove eventually
		self.isOn = None
		self.speedLevel = None
		self.speedIndex = None
		self.speedIndexCount = None
		if self.classID == 9:
			self.typeIsSpeedControl = True
			self.typeSupportsSpeedControl = True		# legacy for RESTful API, remove eventually
			self.typeSupportsOnOff = True
			self.speedLevel = int(round(xmlDev.GetChildElemInt(u"SpeedValue") / 10.0))
			self.speedIndex = int(self.displayRawState)
			self.speedIndexCount = xmlDev.GetChildElemInt(u"NumDiscreteVals")
			if self.speedIndex > 0:
				self.isOn = True
			else:
				self.isOn = False

		self.typeIsSensor = False
		self.typeIsLock = False
		self.typeIsOpenClose = False
		self.sensorValue = None
		self.energyMeterCurLoad = None
		self.energyMeterAccumTotal = None
		if self.classID in [1, 2, 3]:
			if self.classID in [1, 2]:
				self.typeSupportsOnOff = True
				self.isOn = xmlDev.GetChildElemBool(u"IsOn")

				propertiesDict = xmlDev.GetChildElemNode(u"Properties")
				self.typeIsLock = propertiesDict.GetChildElemBool(u"IsLockSubType")
				self.typeIsOpenClose = propertiesDict.GetChildElemBool(u"IsOpenCloseSubType")
			elif self.classID == 3:			# is sensor
				self.typeIsSensor = True
				statesDict = xmlDev.GetChildElemNode(u"States")
				propertiesDict = xmlDev.GetChildElemNode(u"Properties")
				self.typeSupportsOnOff = propertiesDict.GetChildElemBool(u"SupportsOnState")
				self.typeSupportsSensorValue = propertiesDict.GetChildElemBool(u"SupportsSensorValue")
				if self.typeSupportsOnOff:
					self.isOn = xmlDev.GetChildElemBool(u"IsOn")
				if self.typeSupportsSensorValue:
					self.sensorValue = statesDict.GetChildElemStr(u"sensorValue.ui")
					if self.sensorValue is None:
						self.sensorValue = statesDict.GetChildElemStr(u"sensorValue")

				numEnergyInputs = propertiesDict.GetChildElemInt(u"NumEnergyInputs")
				if numEnergyInputs > 0:		# is energy meter sensor
					self.typeSupportsEnergyMeter = True
					self.typeSupportsOnOff = False	# don't really treat as an ON/OFF type
					self.energyMeterCurLoad = statesDict.GetChildElemStr(u"energyInput1.ui")
					self.energyMeterAccumTotal = statesDict.GetChildElemStr(u"accumEnergyTotal1.ui")

					timeBase = propertiesDict.GetChildElemInt(u"AccumEnergyBaseTime1")
					if timeBase > 0:
						self.energyMeterAccumTotal += " used since "
						self.energyMeterAccumTotal += unicode(time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(timeBase + _kEpochOffset)), "utf-8")

		####################
		self.brightness = None
		if self.classID == 2:
			self.typeIsDimmer = True
			self.typeSupportsDim = True			# legacy for RESTful API, remove eventually
			self.brightness = int(round(xmlDev.GetChildElemInt(u"BrightValue") / 10.0))
			self.displayLongState = self.displayLongState + u"%"
		else:
			self.typeIsDimmer = False
			self.typeSupportsDim = False		# legacy for RESTful API, remove eventually

		####################
		self.lastChanged = lastChanged
		if self.lastChanged > 0:
			self.lastChangedDateStr = unicode(time.strftime("%Y-%m-%d", time.gmtime(self.lastChanged + _kEpochOffset)), "utf-8")
			self.lastChangedTimeStr = unicode(time.strftime("%I:%M:%S %p", time.gmtime(self.lastChanged + _kEpochOffset)), "utf-8")
			utcMungedLastChanged = self.lastChanged + _kEpochOffset
			if time.daylight and time.localtime().tm_isdst:
				utcMungedLastChanged += time.altzone
			else:
				utcMungedLastChanged += time.timezone
			self.lastChangedRFC822 = unicode(time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(utcMungedLastChanged)), "utf-8")
			self.lastChangedRFC3339 = unicode(time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(utcMungedLastChanged)), "utf-8")
		else:
			self.lastChangedDateStr = _kEmptyStr
			self.lastChangedTimeStr = _kEmptyStr
			self.lastChangedRFC822 = _kEmptyStr
			self.lastChangedRFC3339 = _kEmptyStr

		self.displayInUI = displayInUI
		self.hasStateToDisplay = hasStateToDisplay
		self._editableAttrs = [
			"isOn", "brightness",
			"speedIndex",
			"activeZone",
			"setpointCool", "setpointHeat", "hvacCurrentMode", "hvacFanMode",
			"outputBinaryStates"
		]

	####################
	def __cmp__(self, other):
		cpSelf = self				# Compare everything but the lastChanged fields.
		cpSelf.lastChanged = None
		cpSelf.lastChangedDateStr = None
		cpSelf.lastChangedTimeStr = None
		cpSelf.lastChangedRFC822 = None
		cpSelf.lastChangedRFC3339 = None
		cpOther = other
		cpOther.lastChanged = None
		cpOther.lastChangedDateStr = None
		cpOther.lastChangedTimeStr = None
		cpOther.lastChangedRFC822 = None
		cpOther.lastChangedRFC3339 = None
		return cmp(cpSelf.__dict__, cpOther.__dict__)

def DeviceFromXml(xmlDev):
	id = xmlDev.GetChildElemInt(u"ID", 0)
	name = xmlDev.GetChildElemStr(u"Name")
	desc = xmlDev.GetChildElemStr(u"Description")
	classID = xmlDev.GetChildElemInt(u"Class")
	devProtocol = xmlDev.GetChildElemInt(u"Protocol")
	type = xmlDev.GetChildElemStr(u"TypeName")
	typeFlags = xmlDev.GetChildElemInt(u"TypeFlags2")
	versByte = xmlDev.GetChildElemInt(u"VersionByte")					# may be None
	address = xmlDev.GetChildElemInt(u"Address")
	lastChanged = xmlDev.GetChildElemInt(u"LastChanged")
	displayInUI = xmlDev.GetChildElemBool(u"DisplayInRemoteUI", True)	# may be None; defaults to True
	displayRawState = xmlDev.GetChildElemStr(u"DeviceDisplayRawState")
	displayLongState = xmlDev.GetChildElemStr(u"DeviceDisplayLongState")
	hasStateToDisplay = xmlDev.GetChildElemBool(u"HasStateToDisplay")	# may be None

	return Device(id, name, desc, classID, devProtocol, type, typeFlags, versByte, address, lastChanged, displayInUI, displayRawState, displayLongState, hasStateToDisplay, xmlDev)

####################
class ControlPage(DictElem):
	""" Document me. """

	####################
	def __init__(self, id, name, desc, categoryID, hiddenInUI, width, height, color, imageFileName):
		DictElem.__init__(self, u'controlpage')
		self.id = id
		self.name = name
		self.desc = desc

		self.categoryID = categoryID
		self.hiddenInUI = hiddenInUI

		self.width = width
		self.height = height
		self.halfWidth = self.width / 2
		self.halfHeight = self.height / 2

		self.imageFileName = imageFileName		# might be None
		self.imageFilePath = None
		if self.imageFileName:
			self.imageFilePath = u"images/backgrounds/" + self.imageFileName

		self.color = color

	####################
	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def ControlPageFromXml(xmlPage):
	id = xmlPage.GetChildElemInt(u"ID", 0)
	name = xmlPage.GetChildElemStr(u"Name")
	desc = xmlPage.GetChildElemStr(u"Description")
	categoryID = xmlPage.GetChildElemInt(u"Category")
	hiddenInUI = xmlPage.GetChildElemBool(u"HiddenInUI")
	(width, height) = xmlPage.GetChildElemIntPair(u"Size")
	color = xmlPage.GetChildElemHtmlColor(u"Color", u"#FFFFFF")
	imageFileName = xmlPage.GetChildElemStr(u"ImageFileName")			# may be None

	return ControlPage(id, name, desc, categoryID, hiddenInUI, width, height, color, imageFileName)

####################
class PageElem(DictElem):
	####################
	""" Document me. """
	def _GetFontFamily(self, fontID):
		if fontID == 0:
			return u"Arial, Helvetica, sans-serif"
		elif fontID == 1:
			return u"'Arial Black', Gadget, sans-serif"
		elif fontID == 2:
			return u"Verdana, Geneva, sans-serif"
		elif fontID == 3:
			return u"Impact, Charcoal, sans-serif"
		elif fontID == 20:
			return u"Lato Light, sans-serif"
		elif fontID == 21:
			return u"Lato Light Italic, sans-serif"
		elif fontID == 22:
			return u"Lato Medium, sans-serif"
		elif fontID == 23:
			return u"Lato Medium Italic, sans-serif"
		elif fontID == 24:
			return u"Lato Heavy, sans-serif"
		elif fontID == 25:
			return u"Lato Heavy Italic, sans-serif"
		elif fontID == 100:
			return u"Georgia, 'Times New Roman', Times, serif"
		elif fontID == 101:
			return u"'Times New Roman', Times, serif"
		elif fontID == 200:
			return u"'Courier New', Courier, monospace"
		elif fontID == 201:
			return u"Monaco, 'Lucida Console', monospace"
		elif fontID == 300:
			return u"'Comic Sans MS', cursive"
		return u"sans-serif"

	####################
	""" Document me. """
	def __init__(
		self,
		controlType, controlIndex, controlValueRaw, controlValueLong, targetDev, targetVar,
		leftPos, topPos, width, height, rotation, alpha,
		showStateImage, showStateText, imageFileName, imageFileNameNeedsResolver, imageFileURL, imageRefreshDuration, stateTextAlignment, stateTextFont, stateTextPointSize, stateTextColor,
		targetElemID, targetElemName, targetElemSubKey,
		captionName, captionPlacement, captionFont, captionPointSize, captionColor, captionCurWidth, captionCurHeight, captionWraps,
		serverActionClass, clientActionType, actionLinkTarget, actionLinkOpensInNewWindow
	):
		DictElem.__init__(self, u'pageelem')
		self.index = controlIndex
		self.controlType = controlType
		self.controlValueRaw = _kEmptyStr
		self.controlValueLong = _kEmptyStr
		if controlValueRaw:
			self.controlValueRaw = controlValueRaw
		if controlValueLong:
			self.controlValueLong = controlValueLong

		self.targetDev = targetDev
		self.targetVar = targetVar

		self.leftPos = leftPos
		self.topPos = topPos
		self.width = width
		self.height = height

		self.rotation = rotation
		self.alpha = alpha

		self.showStateImage = showStateImage
		self.showStateText = showStateText

		self.imageFileName = imageFileName					# might be None
		self.imageFileNameNeedsResolver = imageFileNameNeedsResolver
		self.imageFileURL = imageFileURL					# might be None
		self.imageRefreshDuration = imageRefreshDuration
		self.imageFilePath = None							# will calculate below

		self.stateTextAlignment = stateTextAlignment
		if self.stateTextAlignment == 0:
			self.stateTextAlignmentName = u"left"
		elif self.stateTextAlignment == 1:
			self.stateTextAlignmentName = u"right"
		else:
			self.stateTextAlignmentName = u"center"

		self.stateTextFont = stateTextFont
		self.stateTextFontFamily = self._GetFontFamily(self.stateTextFont)
		self.stateTextPointSize = stateTextPointSize
		self.stateTextColor = stateTextColor

		self.targetElemID = targetElemID			# might be None
		self.targetElemName = targetElemName		# might be None
		self.targetElemSubKey = targetElemSubKey	# might be None

		self.captionName = captionName				# might be None
		self.captionPlacement = captionPlacement
		self.captionFont = captionFont
		self.captionFontFamily = self._GetFontFamily(self.captionFont)
		self.captionPointSize = captionPointSize
		self.captionColor = captionColor
		self.captionCurWidth = captionCurWidth
		self.captionCurHeight = captionCurHeight
		self.captionWraps = captionWraps

		self.serverActionClass = serverActionClass
		self.clientActionType = clientActionType
		self.actionLinkTarget = actionLinkTarget	# might be None
		if actionLinkOpensInNewWindow:
			self.actionLinkHrefTarget = "target=\"_blank\""
		else:
			self.actionLinkHrefTarget = ""

		self.popupControlPath = None
		if self.controlType == 0:
			pass	# none type
		elif self.controlType == 1:		# --- device state control type ---
			self.popupControlPath = u"popupdevicepage"
			if self.showStateImage and self.imageFileName:
				if self.imageFileNameNeedsResolver:
					self.imageFilePath = iu.GetImagePathWithResolver(self.controlType, self.targetElemName, self.controlValueRaw, self.imageFileName)
				else:
					self.imageFilePath = iu.GetImagePathSimple(self.controlType, self.targetElemName, self.controlValueRaw, self.imageFileName)
		elif self.controlType == 2:		# --- variable state control type ---
			self.popupControlPath = u"editvarpage"
			if self.showStateImage and self.imageFileName:
				if self.imageFileNameNeedsResolver:
					self.imageFilePath = iu.GetImagePathWithResolver(self.controlType, self.targetElemName, self.controlValueRaw, self.imageFileName)
				else:
					self.imageFilePath = iu.GetImagePathSimple(self.controlType, self.targetElemName, self.controlValueRaw, self.imageFileName)
		elif self.controlType == 3:		# --- video stream type ---
			pass
			# support for live video streams some day
		elif self.controlType == 80:	# --- folder control type ---
			if self.targetElemSubKey is None:
				self.targetElemSubKey = -1
		elif self.controlType == 100:	# --- static control type ---
			if self.showStateImage and self.imageFileName:
				if self.imageFileNameNeedsResolver:
					self.imageFilePath = iu.GetImagePathWithResolver(self.controlType, self.targetElemName, self.controlValueRaw, self.imageFileName)
				else:
					self.imageFilePath = iu.GetImagePathSimple(self.controlType, self.targetElemName, self.controlValueRaw, self.imageFileName)
		elif self.controlType == 101:	# --- auto refreshing image control type ---
			self.imageRefreshDuration_local = imageRefreshDuration
			self.imageRefreshDuration_remote = imageRefreshDuration
			if self.imageRefreshDuration_local < 150:
				self.imageRefreshDuration_local = 150		# minumum refresh duration of 0.150 seconds for local access
			if self.imageRefreshDuration_remote < 4000:
				self.imageRefreshDuration_remote = 4000		# minumum refresh duration of 4.000 seconds for remote access
		elif self.controlType == 200:	# --- server status text control type ---
			pass
		elif self.controlType == 201:	# --- server status icon control type ---
			pass
		else:
			pass	# invalid / unknown control type

		# It would be ideal to not have the server pass the caption height or the
		# caption width unless the wrap flag is set, in which case we would need
		# the width.
		#
		# However, we need to have that height value here so we can determine the
		# top offset for center vertical alignment. Ideally, we would use CSS to
		# center the caption within its parent DIV but it doesn't look like there
		# is a good way to do this via CSS that is supported by the major browsers.
		# There are some hacks to accomplish it, but I'd prefer to not use those
		# at this point since different browsers use different techniques.
		#
		# Likewise, not knowing the caption width presents some problems. Although
		# it is possible to use a large default width and get the correct visual
		# layout, we have problems with the transparent buffer space used in the
		# tag eating mouse clicks on other controls. For example, consider this:
		#
		#     ( button 1 )     office light ( light icon )
		#
		# The caption "office light" would have an artificially large width that
		# extends over and on top of button 1, resulting in clicks on button 1
		# not working (or triggering the office light action). By knowing the
		# approximate width of the caption, we can shrink the <div> elems to
		# snugly fit thus avoiding the problem most of the time. 2006.01.26 (mmb)
		#
		if not self.captionWraps:
			# If we are not wrapping (single line of text), then add some buffer
			# to the caption width. This is to avoid having the caption overflow
			# onto the graphic. If we add too much fudge, then the caption elem
			# will be larger than it needs to be and might obscure clicking on
			# other elements. If we add too little, then the caption might overlap
			# onto the graphic (for caption to-the-right-of-image alignment)
			# or might not appear centered. If we knew the font metrics used by
			# the browser exactly matched the Indigo control page editor, then
			# this wouldn't be necessary.
			horzFudge = int(self.captionCurWidth * 0.10)
			if horzFudge > 20:
				horzFudge = 20
			self.captionCurWidth += horzFudge

		halfCaptionCurWidth = int(self.captionCurWidth/2)
		halfCaptionHeight = int(self.captionCurHeight/2)

		halfSelfHeight = int(self.height/2)
		halfSelfWidth = int(self.width/2)

		self.captionPosStyle = u"position: absolute;"
		if self.captionPlacement in [0, 1, 2]:
			self.captionPosStyle += u" left: "+str(-halfCaptionCurWidth + halfSelfWidth)+u"px;"
			self.captionPosStyle += u" width: "+str(self.captionCurWidth)+u"px;"
			if self.captionPlacement == 0:
				self.captionPosStyle += u" bottom: "+str(self.height+_kCaptionBufferSpace)+u"px;"
				self.captionPosStyle += u" text-align: center;"
			elif self.captionPlacement == 1:
				self.captionPosStyle += u" top: "+str(halfSelfHeight-halfCaptionHeight)+u"px;"
				self.captionPosStyle += u" text-align: center;"
			else:	# self.captionPlacement == 2:
				self.captionPosStyle += u" top: "+str(self.height+_kCaptionBufferSpace)+u"px;"
				self.captionPosStyle += u" text-align: center;"
		elif self.captionPlacement in [4, 5]:
			self.captionPosStyle += u" top: "+str(halfSelfHeight-halfCaptionHeight)+u"px;"
			if self.captionPlacement == 4:
				self.captionPosStyle += u" right: "+str(self.width+_kCaptionBufferSpace)+u"px;"
				self.captionPosStyle += u" width: "+str(self.captionCurWidth)+u"px;"
				self.captionPosStyle += u" text-align: right;"
			else:	# self.captionPlacement == 5:
				self.captionPosStyle += u" left: "+str(self.width+_kCaptionBufferSpace)+u"px;"
				self.captionPosStyle += u" width: "+str(self.captionCurWidth)+u"px;"
				self.captionPosStyle += u" text-align: left;"
		else:
			pass	# should never happen

		if not self.captionWraps:
			# Force wrapping to off. If the browser font metrics exactly match, or if the
			# horzFudge added above is enough, then this isn't necessary. But force it off
			# as a fail safe measure -- better to have the caption off-center a bit or to
			# overlap onto its parent image a bit than to have it wrap to a second line.
			self.captionPosStyle += u" white-space: nowrap;"

	####################
	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def PageElemFromXml(xmlPageElem):
	controlType = xmlPageElem.GetChildElemInt(u"ControlType")
	controlIndex = xmlPageElem.GetChildElemInt(u"ServerIndex")

	controlValueRaw = xmlPageElem.GetChildElemStr(u"ValueRaw")				# may be None
	controlValueLong = xmlPageElem.GetChildElemStr(u"ValueLong")			# may be None

	targetDev = None	# will be None unless _kGetPageFlags_EmbedTargetElems was used
	if ic.GetChildElementByTagName(xmlPageElem.xmlObj, u"Device") is not None:
		xmlDev = xmlPageElem.GetChildElemNode(u"Device")
		targetDev = DeviceFromXml(xmlDev)

	targetVar = None	# will be None unless _kGetPageFlags_EmbedTargetElems was used
	if ic.GetChildElementByTagName(xmlPageElem.xmlObj, u"Variable") is not None:
		xmlVar = xmlPageElem.GetChildElemNode(u"Variable")
		targetVar = VariableFromXml(xmlVar)

	(leftPos, topPos) = xmlPageElem.GetChildElemIntPair(u"Position")
	(width, height) = xmlPageElem.GetChildElemIntPair(u"Size")
	rotation = xmlPageElem.GetChildElemInt(u"Rotation")						# may be None
	alpha = xmlPageElem.GetChildElemInt(u"Alpha", 100)						# may be None

	showStateImage = xmlPageElem.GetChildElemBool(u"ShowStateImage")
	showStateText = xmlPageElem.GetChildElemBool(u"ShowStateText")
	imageFileName = xmlPageElem.GetChildElemStr(u"ImageFileName")								# may be None
	imageFileNameNeedsResolver = xmlPageElem.GetChildElemBool(u"ImageFileNameNeedsResolver")	# may be None, defaults to False
	imageFileURL = xmlPageElem.GetChildElemStr(u"ImageFileURL")									# may be None
	imageRefreshDuration = xmlPageElem.GetChildElemInt(u"ImageRefreshDuration", 2000)			# may be None

	stateTextAlignment = xmlPageElem.GetChildElemInt(u"StateTextAlignment", 2)		# defaults to align center
	stateTextFont = xmlPageElem.GetChildElemInt(u"StateTextFontType", 0)	# default font ID is 0 (Arial)
	stateTextPointSize = xmlPageElem.GetChildElemInt(u"StateTextPointSize", 12)				# may be None
	stateTextColor = xmlPageElem.GetChildElemHtmlColor(u"StateTextFontColor", u"#008000")	# may be None

	targetElemID = xmlPageElem.GetChildElemInt(u"TargetElemID")			# may be None
	targetElemName = xmlPageElem.GetChildElemStr(u"TargetElemName")			# may be None
	targetElemSubKey = xmlPageElem.GetChildElemStr(u"TargetElemSubKey")		# may be None

	captionName = xmlPageElem.GetChildElemStr(u"CaptionName")				# may be None
	captionPlacement = xmlPageElem.GetChildElemInt(u"CaptionPlacement", 2)	# may be None
	captionFont = xmlPageElem.GetChildElemInt(u"CaptionFontType", 0)		# default font ID is 0 (Arial)
	captionPointSize = xmlPageElem.GetChildElemInt(u"CaptionPointSize", 12)	# may be None
	captionColor = xmlPageElem.GetChildElemHtmlColor(u"CaptionFontColor", u"#000000")	# may be None
	captionCurWidth = xmlPageElem.GetChildElemInt(u"CaptionCurWidth", 200)	# may be None
	captionCurHeight = xmlPageElem.GetChildElemInt(u"CaptionCurHeight", 12)	# may be None
	captionWraps = xmlPageElem.GetChildElemBool(u"CaptionWraps")			# may be None

	serverActionClass = xmlPageElem.GetChildElemInt(u"ServerActionClass", 0)	# kActionNone is default (0)
	clientActionType = xmlPageElem.GetChildElemInt(u"ClientActionType", 0)	# kActionClientNone is default (0)
	actionLinkTarget = xmlPageElem.GetChildElemStr(u"LinkTarget")			# may be None
	actionLinkOpensInNewWindow = xmlPageElem.GetChildElemBool(u"LinkOpensInNewWindow")

	return PageElem(
		controlType, controlIndex, controlValueRaw, controlValueLong, targetDev, targetVar,
		leftPos, topPos, width, height, rotation, alpha,
		showStateImage, showStateText, imageFileName, imageFileNameNeedsResolver, imageFileURL, imageRefreshDuration, stateTextAlignment, stateTextFont, stateTextPointSize, stateTextColor,
		targetElemID, targetElemName, targetElemSubKey,
		captionName, captionPlacement, captionFont, captionPointSize, captionColor, captionCurWidth, captionCurHeight, captionWraps,
		serverActionClass, clientActionType, actionLinkTarget, actionLinkOpensInNewWindow
	)

####################
class Variable(DictElem):
	####################
	""" Document me. """
	def __init__(self, id, name, value, readOnly, displayInUI):
		DictElem.__init__(self, u'variable')
		self.id = id
		self.name = name
		self.value = _kEmptyStr
		if value:
			self.value = value
		if self.value == _kFalseStr:
			self.isFalse = True
		else:
			self.isFalse = False
		if readOnly == _kTrueStr:
			self.readOnly = True
		else:
			self.readOnly = False
		self.displayInUI = displayInUI
		if self.readOnly:
			self._editableAttrs = []
		else:
			self._editableAttrs = ["value"]

	####################
	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def VariableFromXml(xmlVar):
	id = xmlVar.GetChildElemInt(u"ID", 0)
	name = xmlVar.GetChildElemStr(u"Name")
	value = xmlVar.GetChildElemStr(u"Value")
	readOnly = xmlVar.GetChildElemStr(u"ReadOnly")
	displayInUI = xmlVar.GetChildElemBool(u"DisplayInRemoteUI", True)	# may be None; defaults to True

	return Variable(id, name, value, readOnly, displayInUI)

####################
class InsteonLinkRecord(DictElem):
	####################
	""" Document me. """
	def __init__(self, status, persistent, type, deviceName, address, groupNum, data1, data2, data3):
		DictElem.__init__(self, u'insteonlink')
		self.status = status
		self.persistent = persistent
		self.type = type
		self.deviceName = deviceName
		self.address = address
		self.groupNum = groupNum
		self.data1 = data1
		self.data2 = data2
		self.data3 = data3

	####################
	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def InsteonLinkRecordFromXml(xmlLinks):
	status = xmlLinks.GetChildElemStr(u"Status")
	persistent = xmlLinks.GetChildElemStr(u"Persistent")
	type = xmlLinks.GetChildElemStr(u"Type2")
	deviceName = xmlLinks.GetChildElemStr(u"DeviceNameUI")	# DeviceName
	address = xmlLinks.GetChildElemStr(u"AddressStrUI")		# Address, AddressValUI
	groupNum = xmlLinks.GetChildElemStr(u"GroupNumUI")		# GroupNum
	data1 = xmlLinks.GetChildElemStr(u"Data1UI")			# Data1
	data2 = xmlLinks.GetChildElemStr(u"Data2UI")			# Data2
	data3 = xmlLinks.GetChildElemStr(u"Data3UI")			# Data3
#	xmlLinks.GetChildElemStr(u"RawFlags")
#	xmlLinks.GetChildElemStr(u"ControllerDevFlagsUI")
#	xmlLinks.GetChildElemStr(u"ControllerInstnFlagsUI")
#	xmlLinks.GetChildElemStr(u"ControllerInstnLinksDbFormat")
#	xmlLinks.GetChildElemStr(u"ControllerNumControlGroups")
#	xmlLinks.GetChildElemStr(u"ResponderDevFlagsUI")
#	xmlLinks.GetChildElemStr(u"ResponderInstnFlagsUI")
#	xmlLinks.GetChildElemStr(u"ResponderInstnLinksDbFormat")
#	xmlLinks.GetChildElemStr(u"ResponderNumRespondGroups")

	return InsteonLinkRecord(status, persistent, type, deviceName, address, groupNum, data1, data2, data3)

####################
class InsteonLinkProps(DictElem):
	####################
	""" Document me. """
	def __init__(self, status, persistent):
		DictElem.__init__(self, u'insteonlink')
		self.status = status
		self.persistent = persistent

	####################
	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def InsteonLinkPropsFromXml(xmlLinks):
	status = xmlLinks.GetChildElemStr(u"Status")
	persistent = xmlLinks.GetChildElemStr(u"Persistent")

	return InsteonLinkProps(status, persistent)

####################
class InsteonLinks(DictElem):
	####################
	""" Document me. """
	def __init__(self, syncState, syncDevName, linkGroupNamesDict, linkRecordList):
		DictElem.__init__(self, u'insteonlinks')
		self.syncState = _kEmptyStr
		self.syncDevName = _kEmptyStr
		if syncState:
			self.syncState = syncState
		if syncDevName:
			self.syncDevName = syncDevName

		# Eak. Not very clean way to bring the group names up into the main dict.
		if linkGroupNamesDict and linkGroupNamesDict.xmlObj.childNodes:
			for childNode in linkGroupNamesDict.xmlObj.childNodes:
				if childNode.nodeType == childNode.ELEMENT_NODE:
					setattr(self, childNode.localName, childNode.childNodes[0].data)

		self.linkRecordList = linkRecordList

	####################
	def __cmp__(self, other):
		return cmp(self.__dict__, other.__dict__)

def InsteonLinksFromXml(xmlLinks):

#	respondersVector = linksDict["ResponderList"].GetXmlVector()
#	controllersVector = linksDict["ControllerList"].GetXmlVector()
#	emptiesVector = linksDict["EmptyRecordList"].GetXmlVector()

	syncState = xmlLinks.GetChildElemStr(u"SyncState")
	syncDevName = xmlLinks.GetChildElemStr(u"SyncDevName")
	linkGroupNamesDict = xmlLinks.GetChildElemNode(u"ControllerGroupNames")

	linkRecordList = []
	xmlLinkRecordList = xmlLinks.GetListOfChildElems(u"Link")
	for xmlLink in xmlLinkRecordList:
		linkRecordList.append(InsteonLinkRecordFromXml(xmlLink))

	return InsteonLinks(syncState, syncDevName, linkGroupNamesDict, linkRecordList)

####################
class Database:
	""" Document me. """

	####################
	def __init__(self, allowControlPages, feedsEnabled, isDemoServer, logFunc, debugLogFunc):
		self._allowControlPages = allowControlPages
		self._feedsEnabled = feedsEnabled
		self._isDemoServer = isDemoServer

		self._Log = logFunc
		self._DebugLog = debugLogFunc

	def Destroy(self):
		pass

	####################
	def _CalcGetPageFlags(self, ignorePageElems, ignoreActions, retrieveAsList, embedControlTargetElems):
		pageFlags = _kGetPageFlags_None
		if ignorePageElems:
			pageFlags = pageFlags | _kGetPageFlags_IgnorePageElems
		if ignoreActions:
			pageFlags = pageFlags | _kGetPageFlags_IgnoreActions
		if retrieveAsList:
			pageFlags = pageFlags | _kGetPageFlags_RetrieveAsList
		if embedControlTargetElems:
			pageFlags = pageFlags | _kGetPageFlags_EmbedTargetElems
		pageFlags = pageFlags | _kGetPageFlags_EmbedTargetValues
		return pageFlags

	####################
	def AreControlPagesAllowed(self):
		return self._allowControlPages

	def AreFeedsEnabled(self):
		return self._feedsEnabled

	def IsDemoServer(self):
		return self._isDemoServer

	####################
	def GetActionGroup(self, indigoConn, groupName, ignoreActions=False):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", groupName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		if ignoreActions:
			dataNode.AddStringNode(u"IgnoreActions", _kTrueStr)
		responsePacket = indigoConn.SendRequestPacket(u"GetActionGroup", dataNode=dataNode)

		xmlGroup = responsePacket.xmlNode
		group = ActionGroupFromXml(xmlGroup)

		responsePacket.CleanUp()
		return group

	def GetActionGroups(self, indigoConn, doSort, ignoreActions=False):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		if ignoreActions:
			dataNode.AddStringNode(u"IgnoreActions", _kTrueStr)
		responsePacket = indigoConn.SendRequestPacket(u"GetActionGroupList", dataNode=dataNode)

		groupList = []
		xmlGroupList = responsePacket.xmlNode.GetListOfChildElems(u"ActionGroup")
		for xmlGroup in xmlGroupList:
			groupList.append(ActionGroupFromXml(xmlGroup))

		if doSort:
			groupList[:] = SortByAttr(groupList, "name")

		responsePacket.CleanUp()
		return groupList

	####################
	def GetDevice(self, indigoConn, deviceName):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", deviceName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		responsePacket = indigoConn.SendRequestPacket(u"GetDevice", dataNode=dataNode)

		xmlDev = responsePacket.xmlNode
		device = DeviceFromXml(xmlDev)

		responsePacket.CleanUp()
		return device

	def GetDevices(self, indigoConn, doSort):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		responsePacket = indigoConn.SendRequestPacket(u"GetDeviceList", dataNode=dataNode)

		devList = []
		xmlDevList = responsePacket.xmlNode.GetListOfChildElems(u"Device")
		for xmlDev in xmlDevList:
			devList.append(DeviceFromXml(xmlDev))

		if doSort:
			devList[:] = SortByAttr(devList, "name")

		responsePacket.CleanUp()
		return devList

	####################
	def GetControl(self, indigoConn, pageName, controlIndex, ignoreActions=False, retrieveAsList=False, embedControlTargetElems=False):
		'''Return a single control element by the element's z-index '''
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", pageName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"ControlID", controlIndex)

		pageFlags = self._CalcGetPageFlags(False, ignoreActions, retrieveAsList, embedControlTargetElems)
		dataNode.AddStringNode(u"GetPageFlags", str(pageFlags))

		responsePacket = indigoConn.SendRequestPacket(u"GetControlPageElem", dataNode=dataNode)

		xmlPageElem = responsePacket.xmlNode
		controlElem = PageElemFromXml(xmlPageElem)

		responsePacket.CleanUp()
		return controlElem

	def GetControlPage(self, indigoConn, pageName, folderFilter=0, ignorePageElems=False, ignoreActions=False, retrieveAsList=False, embedControlTargetElems=False):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", pageName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"FilterOnFolderID", str(folderFilter))

		pageFlags = self._CalcGetPageFlags(ignorePageElems, ignoreActions, retrieveAsList, embedControlTargetElems)
		dataNode.AddStringNode(u"GetPageFlags", str(pageFlags))

		responsePacket = indigoConn.SendRequestPacket(u"GetControlPage", dataNode=dataNode)

		xmlPage = responsePacket.xmlNode
		page = ControlPageFromXml(xmlPage)

		controlElemList = []
		if not ignorePageElems:
			xmlControlElemList = xmlPage.GetListOfChildElems(u"PageElem")
			for xmlPageElem in xmlControlElemList:
				controlElemList.append(PageElemFromXml(xmlPageElem))

		# change count was increased by server, return the new value
		changeCount = xmlPage.GetChildElemStr(u"ChangeCount")

		responsePacket.CleanUp()
		return [page, controlElemList, changeCount]

	def GetControlPages(self, indigoConn, doSort, ignorePageElems=False, ignoreActions=False, retrieveAsList=False, embedControlTargetElems=False):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()

		pageFlags = self._CalcGetPageFlags(ignorePageElems, ignoreActions, retrieveAsList, embedControlTargetElems)
		dataNode.AddStringNode(u"GetPageFlags", str(pageFlags))

		responsePacket = indigoConn.SendRequestPacket(u"GetControlPageList", dataNode=dataNode)

		pageList = []
		xmlPageList = responsePacket.xmlNode.GetListOfChildElems(u"ControlPage")
		for xmlPage in xmlPageList:
			pageList.append(ControlPageFromXml(xmlPage))

		if doSort:
			pageList[:] = SortByAttr(pageList, "name")

		responsePacket.CleanUp()
		return pageList

	####################
	def GetVariable(self, indigoConn, variableName):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", variableName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		responsePacket = indigoConn.SendRequestPacket(u"GetVariable", dataNode=dataNode)

		xmlVar = responsePacket.xmlNode
		varElem = VariableFromXml(xmlVar)

		responsePacket.CleanUp()
		return varElem

	def GetVariables(self, indigoConn, doSort):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		responsePacket = indigoConn.SendRequestPacket(u"GetVariableList", dataNode=dataNode)

		varList = []
		xmlVarList = responsePacket.xmlNode.GetListOfChildElems(u"Variable")
		for xmlVar in xmlVarList:
			varList.append(VariableFromXml(xmlVar))

		if doSort:
			varList[:] = SortByAttr(varList, "name")

		responsePacket.CleanUp()
		return varList

	####################
	def GetInsteonLinks(self, indigoConn, deviceName):
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()
		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", deviceName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		# Add optional OnlyForControllerGroupNum argument?
		responsePacket = indigoConn.SendRequestPacket(u"GetLinksForDevice", dataNode=dataNode)

		xmlLinks = responsePacket.xmlNode
		insteonLinks = InsteonLinksFromXml(xmlLinks)

		responsePacket.CleanUp()
		return insteonLinks

	####################
	def GetControlPageStatus(self, indigoConn, pageName, changeCount, folderFilter=0, retrieveAsList=False, embedControlTargetElems=False):
		'''Retrieves an update to a page given the pageName and the current changeCount number'''
		if not (self._allowControlPages or self._feedsEnabled):
			raise ControlDisabled
		indigoConn.CheckConnection()

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", pageName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"ChangeCount", str(changeCount))
		dataNode.AddStringNode(u"FilterOnFolderID", str(folderFilter))

		pageFlags = self._CalcGetPageFlags(False, True, retrieveAsList, embedControlTargetElems)
		dataNode.AddStringNode(u"GetPageFlags", str(pageFlags))

		responsePacket = indigoConn.SendRequestPacket(u"GetControlPageStatus2", dataNode=dataNode)

		xmlNode = responsePacket.xmlNode

		page = None
		controlElemList = None
		if ic.GetChildElementByTagName(xmlNode.xmlObj, u"ControlPage") is not None:
			xmlPage = xmlNode.GetChildElemNode(u"ControlPage")
			page = ControlPageFromXml(xmlPage)

			controlElemList = []
			xmlControlElemList = xmlPage.GetListOfChildElems(u"PageElem")
			for xmlPageElem in xmlControlElemList:
				controlElemList.append(PageElemFromXml(xmlPageElem))
		elif ic.GetChildElementByTagName(xmlNode.xmlObj, u"PageElemList") is not None:
			xmlPageList = xmlNode.GetChildElemNode(u"PageElemList")

			controlElemList = []
			xmlControlElemList = xmlPageList.GetListOfChildElems(u"PageElem")
			for xmlPageElem in xmlControlElemList:
				controlElemList.append(PageElemFromXml(xmlPageElem))
		else:
			pass	# no updates needed

		# change count was increased by server, return the new value
		changeCount = xmlNode.GetChildElemStr(u"ChangeCount")

		# server state: idle, busyN
		busyLevel = xmlNode.GetChildElemInt(u"ServerBusyLevel")
		if busyLevel == 0:
			serverState = u"idle"
		else:
			serverState = u"busy" + str(busyLevel)

		responsePacket.CleanUp()
		return [page, controlElemList, changeCount, serverState]

	####################
	def GetServerInfo(self, indigoConn):
		'''Retrieves information about Indigo Server'''
		indigoConn.CheckConnection()

		responsePacket = indigoConn.SendRequestPacket(u"GetServerInfo")
		xmlNode = responsePacket.xmlNode

		# serverVersion = xmlNode.GetChildElemStr(u"ServerVersion")
		licenseLevel = xmlNode.GetChildElemInt(u"LicenseLevel")
		inTrialMode = xmlNode.GetChildElemBool(u"InTrialMode")
		# databaseName = xmlNode.GetChildElemStr(u"DatabaseName")
		# prismReflector = xmlNode.GetChildElemStr(u"PrismReflector")

		return [licenseLevel, inTrialMode]

	####################
	def DeviceTurnOn(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to turn on device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"TurnOn", dataNode=dataNode)

	def DeviceTurnOff(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to turn off device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"TurnOff", dataNode=dataNode)

	def DeviceToggleOnOff(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to toggle device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"ToggleOnOff", dataNode=dataNode)

	def DeviceSetBrightness(self, indigoConn, device, level, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to set device \"" + device + "\" brightness"
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Amount", str(level))
		indigoConn.SendCommandPacket(u"SetBrightness", dataNode=dataNode)

	####################
	def DeviceSetSpeedIndex(self, indigoConn, device, speedIndex, requestOrigin = None):
		speedIndex = int(speedIndex)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		if speedIndex == 0:
			speedStr = "off"
		elif speedIndex == 1:
			speedStr = "low"
		elif speedIndex == 2:
			speedStr = "medium"
		elif speedIndex == 3:
			speedStr = "high"
		else:
			speedStr = "unknown"
		logStr = "request to set device \"" + device + "\" speed to " + speedStr
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Value", str(speedIndex))
		indigoConn.SendCommandPacket(u"SetSpeedIndex", dataNode=dataNode)

	def DeviceDeltaSpeedIndex(self, indigoConn, device, speedIndexDelta, requestOrigin = None):
		speedIndexDelta = int(speedIndexDelta)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		if speedIndexDelta < 0:
			commandName = u"DecreaseSpeedIndex"
			speedIndexDelta = -speedIndexDelta
			logStr = "request to decrease device \"" + device + "\" speed by " + str(speedIndexDelta)
		else:
			commandName = u"IncreaseSpeedIndex"
			logStr = "request to increase device \"" + device + "\" speed by " + str(speedIndexDelta)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Value", str(speedIndexDelta))
		indigoConn.SendCommandPacket(commandName, dataNode=dataNode)

	####################
	def DeviceSetSprinklerActiveZone(self, indigoConn, device, activeZone, requestOrigin = None):
		activeZone = int(activeZone)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		if activeZone <= 0:
			logStr = "request to turn off all device \"" + device + "\" zones"
			if requestOrigin:
				logStr += " from " + requestOrigin
			self._Log(logStr)

			indigoConn.SendCommandPacket(u"SprinklerStopAll", dataVal=device)
		else:
			logStr = "request to set device \"" + device + "\" active zone to " + str(activeZone)
			if requestOrigin:
				logStr += " from " + requestOrigin
			self._Log(logStr)

			dataNode = ic.XmlNode()
			dataNode.AddStringNode(u"Name", device)
			dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
			dataNode.AddStringNode(u"ZoneIndex", str(activeZone))
			indigoConn.SendCommandPacket(u"SprinklerManualOn", dataNode=dataNode)

	def DeviceSprinklerStopAll(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to turn off all device \"" + device + "\" zones"
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"SprinklerStopAll", dataNode=dataNode)

	def DeviceSprinklerRunPrevious(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to run previous zone schedule for device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"SprinklerRunPrevious", dataNode=dataNode)

	def DeviceSprinklerPauseSchedule(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to pause schedule for device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"SprinklerPauseSchedule", dataNode=dataNode)

	def DeviceSprinklerResumeSchedule(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to resume schedule for device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"SprinklerResumeSchedule", dataNode=dataNode)

	def DeviceSprinklerPrevZone(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to change to previous zone for device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"SprinklerPrevZone", dataNode=dataNode)

	def DeviceSprinklerNextZone(self, indigoConn, device, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to change to next zone for device \"" + device + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"SprinklerNextZone", dataNode=dataNode)

	####################
	def DeviceSetHvacSetpointCool(self, indigoConn, device, setpointVal, requestOrigin = None):
		setpointVal = float(setpointVal)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to set device \"" + device + "\" cool setpoint to " + str(setpointVal)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Value", str(setpointVal))
		dataNode.AddStringNode(u"Type", u"cool")
		indigoConn.SendCommandPacket(u"HVACSetpointSet", dataNode=dataNode)

	def DeviceSetHvacSetpointHeat(self, indigoConn, device, setpointVal, requestOrigin = None):
		setpointVal = float(setpointVal)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to set device \"" + device + "\" heat setpoint to " + str(setpointVal)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Value", str(setpointVal))
		dataNode.AddStringNode(u"Type", u"heat")
		indigoConn.SendCommandPacket(u"HVACSetpointSet", dataNode=dataNode)

	def DeviceDeltaHvacSetpointCool(self, indigoConn, device, setpointDelta, requestOrigin = None):
		setpointDelta = float(setpointDelta)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to change device \"" + device + "\" cool setpoint by " + str(setpointDelta)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Delta", str(setpointDelta))
		dataNode.AddStringNode(u"Type", u"cool")
		indigoConn.SendCommandPacket(u"HVACSetpointDelta", dataNode=dataNode)

	def DeviceDeltaHvacSetpointHeat(self, indigoConn, device, setpointDelta, requestOrigin = None):
		setpointDelta = float(setpointDelta)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to change device \"" + device + "\" heat setpoint by " + str(setpointDelta)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Delta", str(setpointDelta))
		dataNode.AddStringNode(u"Type", u"heat")
		indigoConn.SendCommandPacket(u"HVACSetpointDelta", dataNode=dataNode)

	def DeviceSetHvacCurrentMode(self, indigoConn, device, hvacCurrentMode, requestOrigin = None):
		hvacCurrentMode = int(hvacCurrentMode)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to set device \"" + device + "\" hvac mode to \"" + DeviceThermoCurModeToStr(hvacCurrentMode) + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)
		if hvacCurrentMode < 0:
			return

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Action", str(hvacCurrentMode))
		indigoConn.SendCommandPacket(u"HVACModeSet", dataNode=dataNode)

	def DeviceSetHvacFanMode(self, indigoConn, device, hvacFanMode, requestOrigin = None):
		hvacFanMode = int(hvacFanMode)
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to set device \"" + device + "\" fan mode to \"" + DeviceThermoFanModeToStr(hvacFanMode) + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)
		if hvacFanMode < 0:
			return
		hvacFanMode = hvacFanMode + 7	# THVACActionSetMode used; offset up to kHvacActionSetMode_ProgramAuto

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Action", str(hvacFanMode))
		indigoConn.SendCommandPacket(u"HVACModeSet", dataNode=dataNode)

	####################
	def DeviceSetOutputBinaryStates(self, indigoConn, device, binStates, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		stateIndex = 0
		stateTotalValue = 0
		stateList = binStates.split(",")
		stateStr = ""
		for state in stateList:
			if stateIndex > 0:
				stateStr = stateStr + ", "
			state = state.strip()
			if state == "0":
				stateStr = stateStr + "0"
			elif state == "1":
				stateStr = stateStr + "1"
				stateTotalValue = stateTotalValue + (1 << stateIndex)
			else:
				logStr = "request to set device \"" + device + "\" outputBinaryStates"
				if requestOrigin:
					logStr += " from " + requestOrigin
				self._Log(logStr)
				self._Log("failed -- binary output values must be 0 or 1")
				return
			stateIndex = stateIndex + 1

		logStr = "request to set device \"" + device + "\" outputs to " + stateStr
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"OutputOnFlags", str(stateTotalValue))
		indigoConn.SendCommandPacket(u"IOOutputSetVal", dataNode=dataNode)

	def DeviceSetOutputBinaryState(self, indigoConn, device, index, newVal, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		if newVal == 0:
			logStr = "request to clear device \"" + device + "\" output " + str(index)
		else:
			logStr = "request to set device \"" + device + "\" output " + str(index)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)
		if newVal not in [0, 1]:
			self._Log("failed -- binary output values must be 0 or 1")
			return

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", device)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"OutputIndex", str(index))
		dataNode.AddStringNode(u"OutputVal", str(newVal))
		indigoConn.SendCommandPacket(u"IOOutputSetVal", dataNode=dataNode)

	####################
	def InsteonGroupOnCommand(self, indigoConn, groupNum, doFast, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to send insteon group \"" + str(groupNum) + "\" on command"
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"GroupNumber", str(groupNum))
		if doFast:
			dataNode.AddStringNode(u"GroupAction", u"1")
		else:
			dataNode.AddStringNode(u"GroupAction", u"0")
		indigoConn.SendCommandPacket(u"InsteonGroupCommand", dataNode=dataNode)

	def InsteonGroupOffCommand(self, indigoConn, groupNum, doFast, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to send insteon group \"" + str(groupNum) + "\" off command"
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"GroupNumber", str(groupNum))
		if doFast:
			dataNode.AddStringNode(u"GroupAction", u"3")
		else:
			dataNode.AddStringNode(u"GroupAction", u"2")
		indigoConn.SendCommandPacket(u"InsteonGroupCommand", dataNode=dataNode)

	####################
	def VariableAdd(self, indigoConn, variable, value, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to add variable \"" + variable + "\" with value \"" + value + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", variable)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"Value", value)
		dataNode.AddStringNode(u"ReadOnly", _kFalseStr)
		dataNode.AddStringNode(u"ObjVers", u"2")
		indigoConn.SendCommandPacket(u"AddVariable", dataNode=dataNode)

	def VariableSetValue(self, indigoConn, variable, value, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to set variable \"" + variable + "\" value to \"" + value + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		if self._isDemoServer:	# avoid vandalism of variable values on demo server
			self._Log("running as demo server -- request ignored")
		else:
			dataNode = ic.XmlNode()
			dataNode.AddStringNode(u"Name", variable)
			dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
			dataNode.AddStringNode(u"Value", value)
			indigoConn.SendCommandPacket(u"SetVariableValue", dataNode=dataNode)

	####################
	def GroupExecute(self, indigoConn, group, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to execute action group \"" + group + "\""
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", group)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		indigoConn.SendCommandPacket(u"ExecuteActionGroup", dataNode=dataNode)

	####################
	def DoControl(self, indigoConn, pageName, controlIndex, requestOrigin = None):
		if not self._allowControlPages:
			raise ControlDisabled
		indigoConn.CheckConnection()

		logStr = "request to execute \"" + pageName + "\" control #" + str(controlIndex)
		if requestOrigin:
			logStr += " from " + requestOrigin
		self._Log(logStr)

		dataNode = ic.XmlNode()
		dataNode.AddStringNode(u"Name", pageName)
		dataNode.AddStringNode(u"IgnoreCase", _kTrueStr)
		dataNode.AddStringNode(u"ControlID", str(controlIndex))
		indigoConn.SendCommandPacket(u"ExecuteControlAction", dataNode=dataNode)

	####################
	# Only used when JS isn't available on the client browser (useJS=0). This forces
	# the Web server to block between page redirects to give the server time to update
	# devices, variables, etc.
	def BlockUntilServerIdle(self, indigoConn, maxBlockTime = 60000):
		indigoConn.CheckConnection()

		if self._isDemoServer and maxBlockTime > 150:
			maxBlockTime = 150		# don't bog down demo server

		busyLevel = 1
		blockTotal = 0
		while busyLevel > 0 and blockTotal < maxBlockTime:
			responsePacket = indigoConn.SendRequestPacket(u"GetServerBusyLevel")
			busyLevel = responsePacket.xmlNode.GetElemInt()
			responsePacket.CleanUp()
			# self._DebugLog("busyLevel " + str(busyLevel))
			time.sleep(0.250)
			blockTotal = blockTotal + 250
