####################
# Copyright (c) 2021, Perceptive Automation, LLC. All rights reserved.
# http://www.indigodomo.com
#
# Redistribution of this source file, its binary forms, and images are not allowed.
# We are working on our open source license terms, but at this point the source files,
# its binary forms, and all images cannot be redistributed without written permission
# from both Perceptive Automation, LLC. and Andy Turner, HighEarthOrbit software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################

####################
## IMPORTS
try:
	import cherrypy
	import indigo
except:
	pass

import indigoconn as ic
import indigodb as idb
import indigoutil
from basereqhandler import BaseRequestHandler, kTrueStr, kFalseStr, kEmptyStr, kTextPageStr, kHtmlPageStr, kXmlPageStr, kServerStateErrorStr
from actionsreqhandler import ActionsRequestHandler
from devicesreqhandler import DevicesRequestHandler
from variablesreqhandler import VariablesRequestHandler
from insteonlinksreqhandler import InsteonLinksRequestHandler

import base64
import os
import socket
import string
import sys
import time
import traceback
import urllib2
import json
from copy import copy

####################
## CONSTANTS
_kServerStateBusyStr = u"busy1"
_kServerStateIdleStr = u"idle"

# _local intervals for when the browser connection is direct to the server
_kClientPollIntervalBusy_local = 200			# 0.2 second update when server is busy or client is waiting for a response
_kClientPollIntervalIdle_local = 2000			# 2.0 second update interval when server is idle
_kMaxRefreshingImageSize_local = 1500 * 1024	# limit refreshing images to 1.5M

# _remote intervals for when the browser connection is reflected via indigodomo.net
_kClientPollIntervalBusy_remote = 400			# 0.4 second update when server is busy or client is waiting for a response
_kClientPollIntervalIdle_remote = 4000			# 4.0 second update interval when server is idle
_kMaxRefreshingImageSize_remote = 475 * 1024	# limit refreshing images to 475K

_kShortestPollSpacing = 150				# governor: force poll interval schedules to be at least 0.15 seconds apart
_kMaxPollAdjustment = 6000				# upper limit on governor: don't schedule poll interval out more than 6 seconds, regardless of server load

####################
# This has to match the string defined in CPrismReflect.cpp and prism_reflect.py.
_kPrismReflectorStatusStr = u"connected"

# This has to match the strings defined in CControlPage.cpp (kDynamicDevices_ControlPageStr, etc.)
_kInternalDevicesPageStr = u"_internal_devices_"
_kInternalActionGroupsPageStr = u"_internal_action_groups_"
_kInternalVariablesPageStr = u"_internal_variables_"
_kInternalControlPagesPageStr = u"_internal_control_pages_"

####################
_kConnectedStatusStr = u"connected to Indigo Server"

####################
# Read retry values when reading from file (not HTTP) urls.
_kFileUrlRetryAttempts = 3	# Number of attempts to access a file URL before giving up
_kFileUrlRetryDelay = 0.5	# Delay (in seconds) between retries

####################
# Exceptions
class PluginNotInstalled(Exception):
	pass

class TracebackException(Exception):
	pass

####################
class IndigoRequestHandler(BaseRequestHandler):
	""" Handles HTTP page requests. """
	def __init__(self, logger):
		self._earliestPollTime = 0.0
		BaseRequestHandler.__init__(self, logger)
		# REST object handlers:
		self.actions = ActionsRequestHandler(logger)
		self.devices = DevicesRequestHandler(logger)
		self.variables = VariablesRequestHandler(logger)
		self.insteonlinks = InsteonLinksRequestHandler(logger)

	####################
	# The RestDispatcher doesn't currently handle URL type suffixes on the root
	# objects, so we have to force an internal redirect in those cases here.
	def actions_html(self):
		raise cherrypy.InternalRedirect('/actions/index.html')
	def actions_json(self):
		raise cherrypy.InternalRedirect('/actions/index.json')
	def actions_xml(self):
		raise cherrypy.InternalRedirect('/actions/index.xml')
	def actions_txt(self):
		raise cherrypy.InternalRedirect('/actions/index.txt')
	actions_html.exposed = True
	actions_json.exposed = True
	actions_xml.exposed = True
	actions_txt.exposed = True

	def devices_html(self):
		raise cherrypy.InternalRedirect('/devices/index.html')
	def devices_json(self):
		raise cherrypy.InternalRedirect('/devices/index.json')
	def devices_xml(self):
		raise cherrypy.InternalRedirect('/devices/index.xml')
	def devices_txt(self):
		raise cherrypy.InternalRedirect('/devices/index.txt')
	devices_html.exposed = True
	devices_json.exposed = True
	devices_xml.exposed = True
	devices_txt.exposed = True

	def variables_html(self):
		raise cherrypy.InternalRedirect('/variables/index.html')
	def variables_json(self):
		raise cherrypy.InternalRedirect('/variables/index.json')
	def variables_xml(self):
		raise cherrypy.InternalRedirect('/variables/index.xml')
	def variables_txt(self):
		raise cherrypy.InternalRedirect('/variables/index.txt')
	variables_html.exposed = True
	variables_json.exposed = True
	variables_xml.exposed = True
	variables_txt.exposed = True

	def insteonlinks_html(self):
		raise cherrypy.InternalRedirect('/insteonlinks/index.html')
	def insteonlinks_json(self):
		raise cherrypy.InternalRedirect('/insteonlinks/index.json')
	def insteonlinks_xml(self):
		raise cherrypy.InternalRedirect('/insteonlinks/index.xml')
	def insteonlinks_txt(self):
		raise cherrypy.InternalRedirect('/insteonlinks/index.txt')
	insteonlinks_html.exposed = True
	insteonlinks_json.exposed = True
	insteonlinks_xml.exposed = True
	insteonlinks_txt.exposed = True

	####################
	def _LookUpUiName(self, name):
		if name == _kInternalControlPagesPageStr:
			return u"Control Pages"
		elif name == _kInternalDevicesPageStr:
			return u"Devices"
		elif name == _kInternalActionGroupsPageStr:
			return u"Actions"
		elif name == _kInternalVariablesPageStr:
			return u"Variables"
		else:
			return name

	####################
	def _UpdatePollInterval(self, serverState=None, pageName=None):
		if self._IsRequestFromPrism():
			desiredPollInterval = _kClientPollIntervalIdle_remote			# default to slower updates
			if serverState:
				if serverState == kServerStateErrorStr:
					desiredPollInterval = _kClientPollIntervalIdle_remote	# update slowly if server is reporting an error
				elif serverState != _kServerStateIdleStr:
					desiredPollInterval = _kClientPollIntervalBusy_remote	# update quickly while server is busy
		else:
			desiredPollInterval = _kClientPollIntervalIdle_local			# default to slower updates
			if pageName and pageName[:9] == u"_preview_":
				desiredPollInterval = _kClientPollIntervalBusy_local		# always update quickly for preview pages (good for previewing elem drags)
			elif serverState:
				if serverState == kServerStateErrorStr:
					desiredPollInterval = _kClientPollIntervalIdle_local	# update slowly if server is reporting an error
				elif serverState != _kServerStateIdleStr:
					desiredPollInterval = _kClientPollIntervalBusy_local	# update quickly while server is busy

		# Governor: adjust the poll interval based on the number of outstanding poll requests.
		currentTime = time.time()
		desiredPollTime = currentTime + (desiredPollInterval / 1000.0)
		if desiredPollTime < self._earliestPollTime:
			# Desired poll interval time is too soon. To reduce server load, we'll
			# push the poll interval time out so there is at least _kShortestPollSpacing
			# time between each poll interval.
			intervalAdjustment = 1000.0 * (self._earliestPollTime - desiredPollTime)
			self.logger.debug("-- adjusted poll interval " + str(desiredPollInterval) + " by " + str(intervalAdjustment) + " --")
			desiredPollInterval += intervalAdjustment

		if currentTime < self._earliestPollTime:
			# governor is currently on -- increment earliestPollTime by another poll spacing unit
			self._earliestPollTime += (_kShortestPollSpacing / 1000.0)
			maxEarliestPollTime = currentTime + (_kMaxPollAdjustment / 1000.0)
			if self._earliestPollTime > maxEarliestPollTime:
				# Eak... maybe we were Slashdotted. Never let _earliestPollTime creep further
				# than _kMaxPollAdjustment beyond currentTime. We don't have much choice here
				# except to increase the load on the server.
				self.logger.debug("-- poll interval adjustment capped --")
				self._earliestPollTime = maxEarliestPollTime
		else:
			# governor is currently not needed -- reset earliestPollTime based on currentTime
			self._earliestPollTime = currentTime + (_kShortestPollSpacing / 1000.0)

		return desiredPollInterval

	####################
	def index(self, useJS=kTrueStr):
		"""
		Return Type: HTML page
		"""
		(queryString, queryStringNoJs, queryStringUseJs, useJS) = self._GetQueryStrings(useJS)
		redirectPath = cherrypy.url(qs=queryString, relative='server')
		redirectPathNoJs = cherrypy.url(qs=queryStringNoJs, relative='server')
		redirectPathUseJs = cherrypy.url(qs=queryStringUseJs, relative='server')

		webPluginInfoList = cherrypy.server.legacyWebPluginMgr.GetWebPluginInfoList()

		tmpl = self._GetAndLockGlobalTemplate('index.html')
		try:
			self._FillDefaultTemplateAttrs(tmpl)
			tmpl.allowControlPages = cherrypy.server.indigoDb.AreControlPagesAllowed()
			tmpl.feedsEnabled = cherrypy.server.indigoDb.AreFeedsEnabled()
			tmpl.pluginInfoList = webPluginInfoList
			tmpl.redirectPath = redirectPath
			tmpl.redirectPathNoJs = redirectPathNoJs
			tmpl.redirectPathUseJs = redirectPathUseJs
			tmpl.useJS = useJS
			return tmpl.RenderTemplate()
		finally:
			tmpl.ReleaseLock()
	index.exposed = True

	def indigocommand(self, name=None):
		"""
		Return Type: plain text response
		"""
		cherrypy.response.headers['Content-Type'] = 'text/plain'
		name = name.lower()
		if name == u"restart":
			# Haven't tested this, but I'm pretty sure this command is broken
			# with the migration to the plugin Web Server. Could probably just
			# call the plugin restart command here directly now but would need
			# to QA it.
			#
			# Ideally we would just call indigoutil::RestartSelf() here, but forking
			# and starting a new process from a secondary thread isn't recommended.
			# It almost works, but the spawned process doesn't seem to catch termination
			# signals. Delaying the fork/spawn until after the server has stopped (and
			# when we are in the main thread) fixes this.
			if not getattr(cherrypy, "_restart_webserver_", False):
				self.logger.info("request to restart web server from " + cherrypy.request.remote.better_ip)
				cherrypy._restart_webserver_ = True
				# calling cherrypy.engine.stop() stops the server but doesn't unlock it, so just raise SystemExit
				#	cherrypy.engine.stop()
				#	return u"restarted web server"
				raise SystemExit
			else:
				# We can get duplicate requests from the Browser since the first request
				# isn't fulfilled (because the raise SystemExit). Just return that the
				# shutdown is in progress -- do not raise again (it isn't needed).
				return u"web server is shutting down"
		elif name == u"clearcache":
			self.logger.info("request to clear template file cache from " + cherrypy.request.remote.better_ip)
			self._ClearTemplateCache()
			return u"template file cache cleared"
		return u"command not defined"
	indigocommand.exposed = True

	def indigoajax002(self):
		"""
		Return Type: Javascript file
		"""
		cherrypy.response.headers['Content-Type'] = 'text/javascript'
		tmpl = self._GetAndLockGlobalTemplate('indigoajax.js')
		try:
			self._FillDefaultTemplateAttrs(tmpl)
			tmpl.useUniqueUrlToAvoidCache = self._UseUniqueUrlToAvoidCache()
			return tmpl.RenderTemplate()
		finally:
			tmpl.ReleaseLock()
	indigoajax002.exposed = True

	def indigoutil002(self):
		"""
		Return Type: Javascript file
		"""
		cherrypy.response.headers['Content-Type'] = 'text/javascript'
		tmpl = self._GetAndLockGlobalTemplate('indigoutil.js')
		try:
			self._FillDefaultTemplateAttrs(tmpl)
			tmpl.useUniqueUrlToAvoidCache = self._UseUniqueUrlToAvoidCache()
			return tmpl.RenderTemplate()
		finally:
			tmpl.ReleaseLock()
	indigoutil002.exposed = True

	def prismstatus(self, *args, **kwargs):
		"""
		Return Type: plain text response
		"""
		cherrypy.response.headers['Content-Type'] = 'text/plain'
		self._ForceHeadersToNeverCache()
		return _kPrismReflectorStatusStr
	prismstatus.exposed = True

	def message(self, plugin_id, action_id, **kwargs):
		'''
		This handler will pass through whatever is in the POST or GET call to the specified plugin's action_id. The
		incoming URL would be:

			https://someserver.indigodomo.net/message/plugin_id/action_id/?any=extra&params=here
		:param plugin_id: id of the destination plugin
		:param action_id: id of the destination plugin's action
		:param kwargs: query args passed to the message handler will be inserted here
		:return: JSON with an appropriate status codes
				 200 - message was processed successfully and the reply is what was returned from the action
				 501 - plugin isn't installed or action wasn't defined, reply is JSON dict with "error" key
				       containing either "invalid_plugin" or "invalid_action" and a description
				 503 - service unavailable is returned if the plugin is disabled, hoping that it's a temporary issue
				       "error" key in JSON is "plugin_disabled"
				 500 - unknown error is a catch all for any unhandled exceptions
		'''
		self.logger.debug("message handler called for plugin_id: {}, action_id: {}".format(plugin_id, action_id))
		request = cherrypy.request
		if request.method not in ["GET", "POST"]:
			# We only support GET and POST methods to this URL
			raise cherrypy.HTTPError(405, 'Method Not Allowed')

		self._ForceHeadersToNeverCache()
		# We want to pass through the headers, but need to remove the authorization header for security purposes.
		sanitized_headers = copy(request.headers)
		sanitized_headers.pop("Authorization", None)
		props_dict = {
			"incoming_request_method": request.method,
			"headers": sanitized_headers,
		}
		if request.method == "POST":
			# Because N:V pairs in a POST body will be added to kwargs, we need to distinguish them.
			if request.body_params:
				url_query_keys = kwargs.viewkeys() ^ request.body_params.viewkeys()
				body_param_dict = dict()
				for k, v in request.body_params.items():
					body_param_dict[k] = v
				props_dict["body_params"] = body_param_dict
			else:
				url_query_keys = kwargs.viewkeys()
		else:
			url_query_keys = kwargs.keys()
		# Add any query params that were on the URL line
		url_query_args = dict()
		for k in url_query_keys:
			url_query_args[k] = kwargs[k]
		if len(url_query_args):
			props_dict["url_query_args"] = url_query_args
		# Try to get the body from the request - this will fail for a GET, but it will allow a wider range of usages.
		try:
			request_body = request.body.read()
			if len(request_body):
				props_dict["request_body"] = request_body
		except:
			# Assume they didn't pass any body in the POST intentionally and will just use the url params to do whatever
			# processing is necessary.
			pass
		# Call the plugin action
		try:
			self.logger.debug("Calling plugin action {}".format(action_id))
			plugin = indigo.server.getPlugin(plugin_id)
			cherrypy.response.status = 200
			cherrypy.response.headers['Content-Type'] = 'text/json'
			reply_message = plugin.executeAction(action_id, props=props_dict, waitUntilDone=True)
			if isinstance(reply_message, indigo.Dict):
				reply_dict = dict(reply_message)
				if "headers" in reply_dict:
					for k, v in reply_message["headers"].items():
						cherrypy.response.headers[k] = v
				try:
					# just a sanity check that it's actually a numeric
					cherrypy.response.status = int(reply_dict.get("status", 200))
				except:
					# if it's a not a numeric, log a warning but return a 200
					self.logger.warn("'{}' is not a valid HTTP status code, replying with a 200 instead. The plugin should only return valid HTTP codes for the 'status' key in it's reply dictionary.".format(reply_dict.get("status", "")))
				content = reply_dict.get("content", "")
			else:
				# We're just going to assume that JSON was returned and that's what we'll returned
				content = reply_message
			self.logger.debug("reply message:\n{}".format(content))
			return content if isinstance(content, (str, unicode)) else json.dumps(content)
		except indigo.PluginBase.InvalidAction as exc:
			# This gets caught if the plugin doesn't define an action with the specified ID.
			log_message = "plugin with id '{}' doesn't expose an action with id '{}'".format(
						plugin_id, action_id)
			self.logger.error("message handler failed: {}".format(log_message))
			cherrypy.response.status = 501
			cherrypy.response.headers['Content-Type'] = 'text/json'
			return json.dumps(
				{
					"error": "invalid_action",
					"description": log_message,
					"exception": str(exc)
				}
			)
		except indigo.PluginBase.PluginNotInstalled as exc:
			# This gets caught if the plugin isn't installed
			log_message = "plugin with id '{}' isn't installed".format(plugin_id)
			self.logger.error("message handler failed: {}".format(log_message))
			cherrypy.response.status = 501
			cherrypy.response.headers['Content-Type'] = 'text/json'
			return json.dumps(
				{
					"error": "invalid_plugin",
					"description": log_message,
					"exception": str(exc)
				}
			)
		except indigo.PluginBase.PluginDisabled as exc:
			# This gets caught if the plugin is disabled
			log_message = "plugin with id '{}' is disabled".format(plugin_id)
			self.logger.error("message handler failed: {}".format(log_message))
			cherrypy.response.status = 503
			cherrypy.response.headers['Content-Type'] = 'text/json'
			return json.dumps(
				{
					"error": "plugin_disabled",
					"description": log_message,
					"exception": str(exc)
				}
			)
		except Exception as exc:
			# return a message that shows a more generic exception
			log_message = "attempt to communicate with the plugin id '{}' encountered an unexpected error".format(
					plugin_id)
			self.logger.error("message handler failed: {}".format(log_message))
			exception_string = str(exc) if isinstance(exc, TracebackException) else traceback.format_exc()
			self.logger.error("traceback:\n{}".format(exception_string))
			cherrypy.response.status = 500
			cherrypy.response.headers['Content-Type'] = 'text/json'
			message_dict = {
				"error": "unknown_error",
				"description": log_message,
				"exception": exception_string,
			}
			return json.dumps(message_dict)
	message.exposed = True

	####################
	def servercommand(self, rawxml=None):
		"""
		Return Type: Plain text success/failure response
		Description: Passes raw XML command to connected Indigo Server.
		"""
		if rawxml == None:
			return u'error: no rawxml argument'
		cherrypy.response.headers['Content-Type'] = 'text/plain'
		self._ForceHeadersToNeverCache()
		try:
			cherrypy.server.indigoConn.SendCommandRaw(rawxml)
			return u'sent'
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return u'error: %s' % (str(err.value))
	servercommand.exposed = True

	def serverrequest(self, rawxml=None):
		"""
		Return Type: XML raw response from server
		Description: Passes raw XML request to connected Indigo Server
			and returns the XML response.
		"""
		if rawxml == None:
			return u'error: no rawxml argument'
		cherrypy.response.headers['Content-Type'] = 'text/xml'
		self._ForceHeadersToNeverCache()
		try:
			serverResponse = cherrypy.server.indigoConn.SendRequestRaw(rawxml)
			
			# If this is an initial Touch client connection ("GetServerInfo" reply)
			# then log the client's IP address.
			if serverResponse.find('<Name type="string">GetServerInfo</Name>') >= 0:
				if serverResponse.find('<Type type="string">Response</Type>') >= 0:
					remoteClientIP = cherrypy.request.remote.better_ip
					self.logger.info("Indigo Touch client connected from " + remoteClientIP)
			return serverResponse
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			tb = "Error trying to complete /serverrequest: %s\n\n%s" % (str(err.value),cherrypy._cperror.format_exc())
			cherrypy.log(tb, severity=40)
			# Hack to return an IndigoServer style error packet that has some information about what went wrong.
			return u'<Packet type="dict"><Type type="string">Response</Type><Target type="string">Client</Target><Name type="string">Unknown</Name><Error type="dict"><Type type="string">ServerRequestError</Type><Description type="string">%s</Description></Error></Packet>' % (err.value)
	serverrequest.exposed = True

	def serverinfo(self):
		"""
		Return Type: XML info response
		Description: Returns server version info, database name.
		"""
		try:
			cherrypy.response.headers['Content-Type'] = 'text/xml'
			self._ForceHeadersToNeverCache()

			tmpl = self._GetAndLockGlobalTemplate('_serverinfo_response.xml')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.connState = cherrypy.server.indigoConn.GetConnectionState()
				tmpl.serverDbName = cherrypy.server.indigoConn._serverDbName
				tmpl.serverVersion = cherrypy.server.indigoConn._serverVersion
				tmpl.serverLicenseLevel = cherrypy.server.indigoConn._serverLicenseLevel
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kXmlPageStr, err.value)
	serverinfo.exposed = True

	def resolveimagepath(self, type, name, value, filename):
		cherrypy.response.headers['Content-Type'] = 'text/plain'
		thePath = indigoutil.GetImagePathWithResolver(int(type), name, value, filename)
		return thePath
	resolveimagepath.exposed = True

	####################
	def refreshingimage(self, page, id, nocache=None, _=None):
		try:
			pageElem = cherrypy.server.indigoDb.GetControl(cherrypy.server.indigoConn, page, id, True, False, False)
			if not pageElem.imageFileURL:
				cherrypy.server.indigoConn._RaiseServerError("no image URL found for page control " + str(id))

			imageUrl = pageElem.imageFileURL.strip()
			# imageUrl = 'http://static.indigodomo.com/www/images/indigo_title_graphic.jpg'
			# imageUrl = 'file:///blah.png'

			savedTimeout = socket.getdefaulttimeout()
			try:
				try:
					imageReq = None
					if imageUrl.find("file:/") == 0:	# Correct URL format is file:/// but macOS allows file:/ too.
						# Handle file URLs by allowing a number of failed read attempts (defined by constant _kFileUrlRetryAttempts).
						# This is useful because scripts / plugins might be updating the file in real time while we are trying to
						# read from it (file may not exist temporarily).
						imageReq = urllib2.Request(imageUrl)
						retry = _kFileUrlRetryAttempts	# Number of attempts to access a file URL before giving up
						while retry > 0:
							try:
								result = urllib2.urlopen(imageReq)
								result_content_type = result.info().getheader('Content-Type', '')
								result_content_len = int(result.info().getheader('Content-Length', 0))
								break
							except Exception, err:
								retry -= 1
								if retry == 0:
									# Time to give up and raise an error
									cherrypy.server.indigoConn._RaiseServerError(str(err))
								else:
									time.sleep(_kFileUrlRetryDelay)		# Delay (in seconds) between retries 
					elif imageUrl.find("http://") == 0 or imageUrl.find("https://") == 0:
						# Handle username/password authentication URLs like:
						#		http://user:password@127.0.0.1/camera1
						#		http://forcebasic:user:password@127.0.0.1/camera1
						usernamePassword = imageUrl
						usernamePassword = usernamePassword.replace("https://", "")
						usernamePassword = usernamePassword.replace("http://", "")
						atPos = usernamePassword.find("@")
						slashPos = usernamePassword.find("/")
						if atPos > 0 and (slashPos == -1 or atPos < slashPos):
							usernamePassword = usernamePassword.split("@")[0]
							if usernamePassword.find(":") > 0:
								forcebasic = False
								if usernamePassword.find("forcebasic:") == 0:
									imageUrl = imageUrl.replace("forcebasic:", "", 1)
									usernamePassword = usernamePassword.replace("forcebasic:", "", 1)
									forcebasic = True

								(username, password) = usernamePassword.split(":")
								imageUrl = imageUrl.replace(username + ":", "", 1)
								imageUrl = imageUrl.replace(password + "@", "", 1)
								imageReq = urllib2.Request(imageUrl)
								if forcebasic:
									# For some reason installing a HTTP auth handler into the URL opener
									# does not work with all servers. In those cases stuffing the username
									# and password into the header directly can work better.
									b64userpassword = base64.encodestring('%s:%s' % (username, password))[:-1]
									imageReq.add_header("Authorization", "Basic %s" % b64userpassword)
								else:
									passmgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
									passmgr.add_password(None, imageUrl, username, password)
									authhandler1 = urllib2.HTTPBasicAuthHandler(passmgr)
									authhandler2 = urllib2.HTTPDigestAuthHandler(passmgr)
									opener = urllib2.build_opener(authhandler1, authhandler2)
									urllib2.install_opener(opener)
						if imageReq is None:
							imageReq = urllib2.Request(imageUrl)

						socket.setdefaulttimeout(10)	# don't hold up this thread more than 10 seconds waiting for the image
						result = urllib2.urlopen(imageReq)
						if result == None:
							raise Exception('no result')
						if result.getcode() != 200:
							raise Exception('HTTP error %d' % result.getcode())

						result_content_type = result.info().getheader('Content-Type', '')
						result_content_len = int(result.info().getheader('Content-Length', 0))
					else:
						cherrypy.server.indigoConn._RaiseServerError("page control image URL must start with http://, https://, or file:///")
				except (ic.ConnError, ic.InternalError, ic.ServerError), err:
					cherrypy.server.indigoConn._RaiseServerError(imageUrl + " load error: " + err.value)
				except Exception, err:
					cherrypy.server.indigoConn._RaiseServerError(imageUrl + " load error: " + str(err))
			finally:
				socket.setdefaulttimeout(savedTimeout)

			if result_content_type != '' and result_content_type.find('image') != 0:
				# The MIME type returned might be wrong, so check for the common file extensions.
				basename, ext = os.path.splitext(imageUrl)
				if ext not in ['.gif', '.GIF', '.jpg', '.JPG', '.jpeg', '.JPEG', '.png', '.PNG']:
					cherrypy.server.indigoConn._RaiseServerError(imageUrl + " returned type " + result_content_type + " (image required)")
			cherrypy.response.headers['Content-Type'] = result_content_type
			self._ForceHeadersToNeverCache()

			if self._IsRequestFromPrism():
				maxSize = _kMaxRefreshingImageSize_remote
			else:
				maxSize = _kMaxRefreshingImageSize_local
			if result_content_len > maxSize:
				cherrypy.server.indigoConn._RaiseServerError(imageUrl + " size too large: received " + str(result_content_len) + " / max " + str(maxSize))
			# Alternatively or additionally we could add a time.sleep(N) call here, especially
			# if _IsRequestFromPrism() is true, to throttle back the responses. But ideally
			# such bandwidth throttling would be done at the indigodomo.net level.

			# print 'imageFileURL : ' + imageUrl
			# print 'Header : ' + str(result.info()) + '\n\n'
			# print 'Content-Type : ' + result_content_type
			# print 'Content-Length : ' + str(result_content_len)

			try:
				if result_content_len > 0:	# urllib2 bug will hang on some servers (SecuritySpy) if we don't pass a length
					return result.read(result_content_len)
				else:
					return result.read()
			except Exception, err:
				cherrypy.server.indigoConn._RaiseServerError(imageUrl + " read error: " + str(err))
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kHtmlPageStr)
	refreshingimage.exposed = True

	####################
	def controlpage(self, name=_kInternalControlPagesPageStr, uiName=None, folderFilter=0, asList=kFalseStr, serverStatus=None, useJS=kTrueStr):
		"""
		Return Type: HTML page
		"""
		if not cherrypy.server.indigoDb.AreControlPagesAllowed():
			return self._ReturnControlDisabled(kHtmlPageStr)
		try:
			self._ForceHeadersToNeverCache()

			if name == _kInternalControlPagesPageStr:
				asList = True;
			else:
				if asList.lower() in [kTrueStr, u"yes", u"1"]:
					asList = True
				else:
					asList = False
			(queryString, queryStringNoJs, queryStringUseJs, useJS) = self._GetQueryStrings(useJS)
			redirectPath = cherrypy.url(qs=queryString, relative='server')
			redirectPathNoJs = cherrypy.url(qs=queryStringNoJs, relative='server')
			redirectPathUseJs = cherrypy.url(qs=queryStringUseJs, relative='server')
			(page, pageElemList, changeCount) = cherrypy.server.indigoDb.GetControlPage(cherrypy.server.indigoConn, name, folderFilter, False, True, asList, asList)
			if page.categoryID == 1 and not asList:
				# list type pages can only be shown as lists no matter what the caller requests
				return self._ReturnListOnlyError(kHtmlPageStr)

			if uiName is None or uiName == "":
				uiName = self._LookUpUiName(name)
			uiParentName = None		# for breadcrumb nav when inside a folder
			if folderFilter != 0:
				uiParentName = self._LookUpUiName(name)

			if serverStatus == None:
				serverStatus = _kConnectedStatusStr

			tmpl = self._GetAndLockGlobalTemplate('controlpage.html')
			try:				
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.page = page
				tmpl.pageName = name
				tmpl.pageElemList = pageElemList
				tmpl.uiName = uiName
				tmpl.uiParentName = uiParentName
				tmpl.asList = asList
				tmpl.folderFilter = folderFilter
				tmpl.changeCount = changeCount
				tmpl.serverStatus = serverStatus
				tmpl.serverState = _kServerStateIdleStr
				tmpl.pollInterval = self._UpdatePollInterval(_kServerStateIdleStr, name)
				tmpl.redirectPath = redirectPath
				tmpl.redirectPathNoJs = redirectPathNoJs
				tmpl.redirectPathUseJs = redirectPathUseJs
				tmpl.useJS = useJS
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			# Demo server hack to avoid missing page errors (just go to index):
			if cherrypy.server.is_demo_server and err.value == "requested control page does not exist":
				return self.controlpage()
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kHtmlPageStr)
	controlpage.exposed = True

	def _getupdatedcontrols(self, name=None, changeCount=0, folderFilter=0, asList=kFalseStr, serverState=None, nocache=None, _=None):
		"""
		Return Type: XML ajax-response 
		Description: Get the list of page elements that have updated since 
			the provided "changeCount" parameter.
		Parameters:
			name - Name of the page to get the updates of,
			changeCount - Current revision of the page the client is viewing.
			serverState - 
			nocache - Unique var used to prevent caching by certain browsers, not used by server.
			_ - Hidden parameter used by the Rico Ajax-engine.
		"""
		try:
			cherrypy.response.headers['Content-Type'] = 'text/xml'
			self._ForceHeadersToNeverCache()

			if asList.lower() in [kTrueStr, u"yes", u"1"]:
				asList = True
			else:
				asList = False
			(page, pageElemList, changeCount, newServerState) = cherrypy.server.indigoDb.GetControlPageStatus(cherrypy.server.indigoConn, name, changeCount, folderFilter, asList, asList)

			closeErrorDialog = None
			pollInterval = self._UpdatePollInterval(newServerState, name)

			if newServerState == serverState:
				newServerState = None			# No change in status -- clear to prevent unnecessary swaps in browser.
			else:
				self.logger.debug("-- server state change: " + newServerState + " --")
				if serverState == kServerStateErrorStr:
					closeErrorDialog = True		# old state caused modal error dialog; can now close it

			if page is not None and pageElemList is not None:
				self.logger.debug("-- refreshing entire page --")
				serverStatus = u"page updated"

				tmpl = self._GetAndLockGlobalTemplate('_pageelemlist_response.xml')
				try:
					self._FillDefaultTemplateAttrs(tmpl)
					tmpl.page = page
					tmpl.pageName = name
					tmpl.pageElemList = pageElemList
					tmpl.asList = asList
					tmpl.folderFilter = folderFilter
					tmpl.changeCount = changeCount
					tmpl.reregisterServerStatus = True
					tmpl.serverStatus = serverStatus
					tmpl.serverState = newServerState
					tmpl.pollInterval = pollInterval
					tmpl.closeErrorDialog = closeErrorDialog
					tmpl.redirectPath = u"#"		# not used in our case (useJS must be True)
					tmpl.useJS = True
					return tmpl.RenderTemplate()
				finally:
					tmpl.ReleaseLock()
			elif pageElemList is not None:
				reregisterServerStatus = False
				commonTargetName = None
				for pageElem in pageElemList:
					self.logger.debug("-- refreshing control " + str(pageElem.index) + " --")
					if pageElem.controlType == 1 or pageElem.controlType == 2:
						if not commonTargetName:
							commonTargetName = pageElem.targetElemName
						elif commonTargetName == "***":
							pass						# multiple controls updated with unique target elems
						elif commonTargetName != pageElem.targetElemName:
							commonTargetName = "***"	# multiple controls updated with unique target elems
					elif pageElem.controlType == 200 or pageElem.controlType == 201:
						reregisterServerStatus = True	# must have client reregister server status elems the server status control is being swapped...
				pageName = name
				if commonTargetName and commonTargetName != "***":
					serverStatus = commonTargetName + u" updated"
				else:
					elemCount = len(pageElemList)
					if elemCount == 1:
						serverStatus = u"control updated"
					else:
						serverStatus = str(elemCount) + u" controls updated"

				tmpl = self._GetAndLockGlobalTemplate('_pageelem_response.xml')
				try:
					self._FillDefaultTemplateAttrs(tmpl)
					tmpl.pageName = name
					tmpl.pageElemList = pageElemList
					tmpl.asList = asList
					tmpl.folderFilter = folderFilter
					tmpl.changeCount = changeCount
					tmpl.reregisterServerStatus = reregisterServerStatus
					tmpl.serverStatus = serverStatus
					tmpl.serverState = newServerState
					tmpl.pollInterval = pollInterval
					tmpl.closeErrorDialog = closeErrorDialog
					tmpl.redirectPath = u"#"		# not used in our case (useJS must be True)
					tmpl.useJS = True
					return tmpl.RenderTemplate()
				finally:
					tmpl.ReleaseLock()
			else:
				self.logger.debug("-- no page changes --")
				serverStatus = None

				tmpl = self._GetAndLockGlobalTemplate('_status_response.xml')
				try:
					self._FillDefaultTemplateAttrs(tmpl)
					tmpl.changeCount = changeCount
					tmpl.reregisterServerStatus = False
					tmpl.serverStatus = serverStatus
					tmpl.serverState = newServerState
					tmpl.pollInterval = pollInterval
					tmpl.closeErrorDialog = closeErrorDialog
					tmpl.redirectPath = u"#"		# not used in our case (useJS must be True)
					tmpl.useJS = True
					return tmpl.RenderTemplate()
				finally:
					tmpl.ReleaseLock()

		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kXmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kXmlPageStr)
		except:
			traceback.print_exc()
	_getupdatedcontrols.exposed = True

	def _displayMessage(self, title, details=None):
		"""
		Return Type: HTML page 
		Description: Displays a message on the page
		Parameters:
			title - Primary message to display.
			details - Any pertinent message details.
		"""
		tmpl = self._GetAndLockGlobalTemplate('error.html')
		try:
			self._FillDefaultTemplateAttrs(tmpl)
			tmpl.errorMsg = title
			tmpl.errorDetails = details
			return tmpl.RenderTemplate()
		finally:
			tmpl.ReleaseLock()
	_displayMessage.exposed = True
	
	def _docontrol(self, page, id=0, redirectPath=None, nocache=None, _=None):
		"""
		Return Type: XML ajax-response containing status (IFF redirectPath not specified)
		Description: Tell Indigo Server to execute a control on a specific page.
		"""
		try:
			cherrypy.server.indigoDb.DoControl(cherrypy.server.indigoConn, page, id, cherrypy.request.remote.better_ip)

			if redirectPath and redirectPath != "":
				cherrypy.server.indigoDb.BlockUntilServerIdle(cherrypy.server.indigoConn, cherrypy.server.max_control_click_wait)
				raise cherrypy.HTTPRedirect(redirectPath, 303)
			else:
				cherrypy.response.headers['Content-Type'] = 'text/xml'
				self._ForceHeadersToNeverCache()
	
				tmpl = self._GetAndLockGlobalTemplate('_status_response.xml')
				try:
					self._FillDefaultTemplateAttrs(tmpl)
					tmpl.reregisterServerStatus = False
					tmpl.serverStatus = None
					tmpl.serverState = _kServerStateBusyStr
					tmpl.pollInterval = self._UpdatePollInterval(_kServerStateBusyStr, page)
					tmpl.closeErrorDialog = False
					return tmpl.RenderTemplate()
				finally:
					tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kXmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kXmlPageStr)
	_docontrol.exposed = True

	def _docontroldevice(self, device, command, value=None, nocache=None, _=None):
		"""
		Return Type: XML ajax-response containing status
		Description: Control a device by turning it on, off, or setting dimmable value
		Parameters:
			device - Name of the device to control (Indigo Device Name).
			command - Command to execute on device (turnon, turnoff, dim).
			value - If the command is "dim", this should be the brightness value, 0-100% brightness.
			nocache - Unique var used to prevent caching by certain browsers, not used by server.
			_ - Hidden parameter used by the Rico Ajax-engine.
		"""
		try:
			cherrypy.response.headers['Content-Type'] = 'text/xml'
			self._ForceHeadersToNeverCache()

			if command == u"turnon":
				cherrypy.server.indigoDb.DeviceTurnOn(cherrypy.server.indigoConn, device, cherrypy.request.remote.better_ip)
			elif command == u"turnoff":
				cherrypy.server.indigoDb.DeviceTurnOff(cherrypy.server.indigoConn, device, cherrypy.request.remote.better_ip)
			elif command == u"dim":
				cherrypy.server.indigoDb.DeviceSetBrightness(cherrypy.server.indigoConn, device, value, cherrypy.request.remote.better_ip)

			tmpl = self._GetAndLockGlobalTemplate('_status_response.xml')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.reregisterServerStatus = False
				tmpl.serverStatus = None
				tmpl.serverState = _kServerStateBusyStr
				tmpl.pollInterval = self._UpdatePollInterval(_kServerStateBusyStr)
				tmpl.closeErrorDialog = False
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kXmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kXmlPageStr)
	_docontroldevice.exposed = True

	####################
	def popupdevicepage(self, name, redirectPage=None, redirectPath=None, useJS=kTrueStr):
		"""
		Return Type: HTML page
		"""
		if not cherrypy.server.indigoDb.AreControlPagesAllowed():
			return self._ReturnControlDisabled(kHtmlPageStr)
		try:
			self._ForceHeadersToNeverCache()
			(queryString, queryStringNoJs, queryStringUseJs, useJS) = self._GetQueryStrings(useJS)
			if not redirectPath:
				if redirectPage:
					redirectPath = cherrypy.request.script_name + "/controlpage?name=" + redirectPage + "&useJS=" + useJS
				else:
					redirectPath = cherrypy.url(qs=queryString, relative='server')

			targetDev = cherrypy.server.indigoDb.GetDevice(cherrypy.server.indigoConn, name)

			tmpl = self._GetAndLockGlobalTemplate('popupdevicepage.html')
			try:				
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.targetDev = targetDev
				tmpl.uiName = name
				tmpl.redirectPath = redirectPath
				tmpl.useJS = useJS
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			# Demo server hack to avoid missing page errors (just go to index):
			if cherrypy.server.is_demo_server and err.value == "requested device element does not exist":
				return self.controlpage()
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kHtmlPageStr)
	popupdevicepage.exposed = True

	####################
	def editvarpage(self, name, redirectPage=None, redirectPath=None, useJS=kTrueStr):
		"""
		Return Type: HTML page
		"""
		if not cherrypy.server.indigoDb.AreControlPagesAllowed():
			return self._ReturnControlDisabled(kHtmlPageStr)
		try:
			self._ForceHeadersToNeverCache()
			(queryString, queryStringNoJs, queryStringUseJs, useJS) = self._GetQueryStrings(useJS)
			if not redirectPath:
				if redirectPage:
					redirectPath = cherrypy.request.script_name + "/controlpage?name=" + redirectPage + "&useJS=" + useJS
				else:
					redirectPath = cherrypy.url(qs=queryString, relative='server')

			elem = cherrypy.server.indigoDb.GetVariable(cherrypy.server.indigoConn, name)

			tmpl = self._GetAndLockGlobalTemplate('editvarpage.html')
			try:				
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.elem = elem
				tmpl.uiName = name
				tmpl.redirectPath = redirectPath
				tmpl.useJS = useJS
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			# Demo server hack to avoid missing page errors (just go to index):
			if cherrypy.server.is_demo_server and err.value == "requested variable element does not exist":
				return self.controlpage()
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kHtmlPageStr)
	editvarpage.exposed = True
	
	####################
	def basicpage(self, miniUI=kFalseStr, width=kEmptyStr, serverStatus=None):
		"""
		Return Type: HTML page
		"""
		if not cherrypy.server.indigoDb.AreControlPagesAllowed():
			return self._ReturnControlDisabled(kHtmlPageStr)
		try:
			self._ForceHeadersToNeverCache()

			redirectPath = cherrypy.url(qs=cherrypy.request.query_string, relative='server')
			devList = cherrypy.server.indigoDb.GetDevices(cherrypy.server.indigoConn, True)
			actionGroupList = cherrypy.server.indigoDb.GetActionGroups(cherrypy.server.indigoConn, True, True)
			variableList = cherrypy.server.indigoDb.GetVariables(cherrypy.server.indigoConn, True)
			if serverStatus == None:
				serverStatus = _kConnectedStatusStr

			if miniUI.lower() in [kTrueStr, u"yes", u"1"]:
				tmpl = self._GetAndLockGlobalTemplate('basicpage_mini.html')
			else:
				tmpl = self._GetAndLockGlobalTemplate('basicpage.html')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.redirectPath = redirectPath
				tmpl.serverStatus = serverStatus
				tmpl.devList = devList
				tmpl.actionGroupList = actionGroupList
				tmpl.variableList = variableList
				tmpl.width = width
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kHtmlPageStr)
	basicpage.exposed = True

	def rss(self, serverStatus=None):
		"""
		Return Type: XML rss 2.0 feed, HTML page if there is an error
		"""
		if not cherrypy.server.indigoDb.AreFeedsEnabled():
			return self._ReturnFeedsDisabled(kHtmlPageStr)
		try:
			feedTime = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())
			devList = cherrypy.server.indigoDb.GetDevices(cherrypy.server.indigoConn, False)
			actionGroupList = cherrypy.server.indigoDb.GetActionGroups(cherrypy.server.indigoConn, False, True)
			if serverStatus == None:
				serverStatus = _kConnectedStatusStr

			cherrypy.response.headers['Content-Type'] = 'text/xml'
			tmpl = self._GetAndLockGlobalTemplate('rss.xml')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.feedTime = feedTime
				tmpl.devList = devList
				tmpl.actionGroupList = actionGroupList
				tmpl.serverStatus = serverStatus
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnFeedsDisabled(kHtmlPageStr)
	rss.exposed = True

	def atom(self, serverStatus=None):
		"""
		Return Type: XML atom feed, HTML page if there is an error
		"""
		if not cherrypy.server.indigoDb.AreFeedsEnabled():
			return self._ReturnFeedsDisabled(kHtmlPageStr)
		try:
			feedTime = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
			devList = cherrypy.server.indigoDb.GetDevices(cherrypy.server.indigoConn, False)
			actionGroupList = cherrypy.server.indigoDb.GetActionGroups(cherrypy.server.indigoConn, False, True)
			if serverStatus == None:
				serverStatus = _kConnectedStatusStr

			cherrypy.response.headers['Content-Type'] = 'text/xml'
			tmpl = self._GetAndLockGlobalTemplate('atom.xml')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.feedTime = feedTime
				tmpl.devList = devList
				tmpl.actionGroupList = actionGroupList
				tmpl.serverStatus = serverStatus
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnFeedsDisabled(kHtmlPageStr)
	atom.exposed = True
	
	def devicehistory(self, name):
		"""
		Return Type: HTML page
		"""
		if not cherrypy.server.indigoDb.AreControlPagesAllowed():
			return self._ReturnControlDisabled(kHtmlPageStr)
		try:
			self._ForceHeadersToNeverCache()

			redirectPath = cherrypy.url(qs=cherrypy.request.query_string, relative='server')
			device = cherrypy.server.indigoDb.GetDevice(cherrypy.server.indigoConn, name)

			tmpl = self._GetAndLockGlobalTemplate('devicehistory.html')
			try:
				self._FillDefaultTemplateAttrs(tmpl)
				tmpl.redirectPath = redirectPath
				tmpl.device = device
				return tmpl.RenderTemplate()
			finally:
				tmpl.ReleaseLock()
		except (ic.ConnError, ic.InternalError, ic.ServerError), err:
			return self._ReturnError(kHtmlPageStr, err.value)
		except idb.ControlDisabled:
			return self._ReturnControlDisabled(kHtmlPageStr)
	devicehistory.exposed = True
