####################
# Copyright (c) 2021, Perceptive Automation, LLC. All rights reserved.
# http://www.indigodomo.com
#
# Redistribution of this source file, its binary forms, and images are not allowed.
# We are working on our open source license terms, but at this point the source files,
# its binary forms, and all images cannot be redistributed without written permission
# from both Perceptive Automation, LLC. and Andy Turner, HighEarthOrbit software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################

####################
## IMPORTS
try:
	import indigo
except:
	pass

import cherrypy
from cherrypy import _cperror

import os
import sys
import time
import datetime

####################
####################
def RestartSelf():
	my_sys_executable = sys.executable
	if not my_sys_executable or my_sys_executable == "/":
		# For some unknown reason sys.executable contains just "/" when IndigoServer
		# launches IWS. This is new to OS X 10.5 (Python 2.6) and does NOT occur if
		# you launch IWS from the Terminal. We just hack around the problem for now:
		my_sys_executable = "/usr/bin/python"

	args = sys.argv[:]
	args.insert(0, my_sys_executable)
	if sys.platform == "win32":
		args = ['"%s"' % arg for arg in args]

	pid = os.fork()
	if pid:		# we are the parent; reap children and exit immediately
		try:
			os.waitpid(pid, os.WNOHANG)
		except:
			pass
		os._exit(0)
	else:		# we are the child; fork one more time before we execv()
		os.setsid()
		os.umask(0)
		pid = os.fork()
		if pid:		# we are the parent (but child of original parent)
			try:
				os.waitpid(pid, os.WNOHANG)
			except:
				pass
			os._exit(0)

		# we want to close all open file descriptors, except stdio
		for fd in range(0, os.sysconf('SC_OPEN_MAX')):
			try:
				if fd == sys.stdin.fileno():
					continue
				if fd == sys.stdout.fileno():
					continue
				# Yuck -- _cpwsgi.py breaks stderr by setting it to NullWriter which doesn't
				# define fileno(), so check for its existance first.
				if hasattr(sys.stderr, "fileno") and fd == sys.stderr.fileno():
					continue

				# print "closing fd " + str(fd)
				os.close(fd)
			except OSError:
				pass

		os.execv(my_sys_executable, args)
		os._exit(0)		# we never get here

####################
def DebugLog(logMessage):
	try:
		logFunc = cherrypy.log
		if logFunc:
			logFunc("INDIGO_LOG " + logMessage.encode('ascii', 'replace'))
	except:
		pass	# should never happen -- ignore if it does since we are just logging

def WriteLog(logMessage):
	# Logs to CherryPy AND to Indigo Server if server connection is good
	try:
		logFunc = cherrypy.log
		if logFunc:
			logFunc("INDIGO_LOG " + logMessage.encode('ascii', 'replace'))
	except:
		pass	# should never happen -- ignore if it does since we are just logging
	try:
		if cherrypy.server.indigoConn:
			cherrypy.server.indigoConn.ServerWriteLog(logMessage)
	except:
		pass	# should never happen -- ignore if it does since we are just logging

####################
kHtmlEscapeTable = {
	"&" : "&amp;",
	'"' : "&quot;",
	"'" : "&apos;",		# Yuck: entity was not supported by HTML 4 so IE 5.5/6 Win will display &apos;
	">" : "&gt;",
	"<" : "&lt;"
}
kXmlEscapeTable = {
	"&" : "&amp;",
	">" : "&gt;",
	"<" : "&lt;"
}

def EscapeForHTML(rawUnicode):
	list = []
	for c in rawUnicode:
		list.append(kHtmlEscapeTable.get(c, c))
	return "".join(list)

def EscapeForXML(rawUnicode):
	list = []
	for c in rawUnicode:
		list.append(kXmlEscapeTable.get(c, c))
	return "".join(list)

####################
####################
def ModAuthHeader():
	# If there's no Authorization header, check to see if there's an api_key in the query keys
	# and stuff it into the headers.
	if not cherrypy.request.headers.has_key("Authorization") and cherrypy.request.params.has_key("api_key"):
		cherrypy.request.headers['Authorization'] = "Bearer " + cherrypy.request.params["api_key"]
		del cherrypy.request.params["api_key"]

def ModXForwardedHostForPrism():
	# IndigoDomo.net rewrites the request to an internal port number, which maps to
	# a ssh tunnel to the user's Mac:
	#
	#		  User URL: http://mattb.indigodomo.net
	#		Mapped URL: http://indigodomo.net:9812
	#
	# We must use the X-Forwarded-Host to get the true original URL requested,
	# since the Host key has the rewritten URL. The rewritten URL won't work
	# since indigodomo.net has those port number blocked for remote access (they
	# work for local access which is how the rewrite mod can work).
	#
	# We also use the X-Forwarded-For IP/host name if it exists for calculating
	# a more useful client IP address.
	if cherrypy.request.headers.has_key("X-Forwarded-Host"):
		old_base = cherrypy.request.base
		new_base = cherrypy.request.headers.get("X-Forwarded-Host")
		if new_base.find("://") == -1:	# copy the http:// or https:// from original
			new_base = old_base[:old_base.find("://") + 3] + new_base
		cherrypy.request.base = new_base

	if cherrypy.request.headers.has_key("X-Forwarded-For"):
		xff = cherrypy.request.headers.get("X-Forwarded-For")
		cherrypy.request.remote.better_ip = xff.split(',')[-1].strip()
	else:
		cherrypy.request.remote.better_ip = cherrypy.request.remote.ip

####################
def ModHttpServerName():
	cherrypy.response.headers['Server'] = 'IndigoWebServer/5.0'

####################
def IsLeapYear(year):
	if year % 4 != 0:
		return False
	if year >= 1582 and (year % 100) == 0 and (year % 400) != 0:
		return False
	return True

def isFloatCompat(something):
	try:
		if type(something) == float:
			return True
		float(unicode(something))
		return True
	except TypeError:
		return False
	except ValueError:
		return False

def isIntCompat(something):
	try:
		if type(something) == int:
			return True
		int(unicode(something))
		return True
	except TypeError:
		return False
	except ValueError:
		return False

def CalcNumDaysInMonth(month, year):
	if year <= 1582 or month < 1 or month > 12:
		raise Exception, u"year value is out-of-range"

	if month == 1 or month == 3 or month == 5 or month == 7 or month == 8 or month == 10 or month == 12:
		return 31
	elif month == 4 or month == 6 or month == 9 or month == 11:
		return 30
	else:	# if (month == 2)
		if IsLeapYear(year):
			return 29
		else:
			return 28

def GetNthString(val):
	val = int(val)
	nthStr = str(val)

	lastTwoDigits = val % 100
	if lastTwoDigits >= 11 and lastTwoDigits <= 13:
		nthStr += u"th"			# special case override for 11th, 12th, 13th
	else:
		lastDigit = val % 10
		if lastDigit == 1:
			nthStr += u"st"		# 1st, 21st, 31st, 41st, ..., 101st, 121st, 131st, ...
		elif lastDigit == 2:
			nthStr += u"nd"		# 2nd, 22nd, 32nd, 42nd, ..., 102nd, 122nd, 132nd, ...
		elif lastDigit == 3:
			nthStr += u"rd"		# 3rd, 23rd, 33rd, 43rd, ..., 103rd, 123rd, 133rd, ...
		else:
			nthStr += u"th"		# everything else
	return nthStr
	
def GetNextIntegerStep(integer, step):
	rem = integer % step
	mid = step/2
	if (rem <= mid):
		return integer - rem
	return integer + (step - rem)

####################
def ConvertUtcToLocal(utcDateTime):
	# Is this a good way to convert a datetime.datetime object from UTC to local?
	if time.daylight:
		offsetSeconds = time.altzone
	else:
		offsetSeconds = time.timezone
	timeDelta = datetime.timedelta(seconds=offsetSeconds)
	localDateTime = utcDateTime - timeDelta
	return localDateTime

def ConvertDateTimeToSecs(dt):
	return time.mktime(dt.timetuple())

def ConvertSqlDateTimeStrToDateTime(dtStr):
	dt = datetime.datetime(*(time.strptime(dtStr[0:19], "%Y-%m-%d %H:%M:%S")[0:6]))
	return dt

def GetCurrentTimeTruple():
	return time.localtime()

####################
# A faster version of GetImagePathWithResolver() that only handles binary images.
####################
def GetImagePathSimple(controlType, name, value, filename):
	fullPath = u"images/controls/"

	# Identify the control type
	if controlType == 1:		# It's a device
		fullPath += u"devices/" + filename
		fullPathNoExt = fullPath[:-4]
		fileExt = fullPath[-4:]
		if name and value:
			# 2011.09.13 IS5.0.0b9/IT1.8 Added additional boolean types to test: no, false, 0.0.
			if value.lower() not in [u"", u"false", u"no", u"off", u"0", u"0.0", u"paused"]:
				onPath = fullPathNoExt + u"+on" + fileExt
				return onPath
#
# Although I like the idea of this additional check (before returning onPath),
# mirroring the logic in Indigo Touch isn't possible since it cannot check
# to see if the file exists before requesting it (and performing multiple
# HTTP requests is not good for performance). So for consistency, we do not
# do this check here either; meaning that device images should always have
# a +on variant if they use the simple heuristic.
#
#				USER_STATIC_FOLDER_PATH = cherrypy.config.get("user_static_folder_path")
#				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, onPath)):
#					return onPath
		return fullPath
	elif controlType == 2:		# It's a variable
		fullPath += u"variables/" + filename
		fullPathNoExt = fullPath[:-4]
		fileExt = fullPath[-4:]
		if name and value:
			if value.lower() not in [u"", u"false", u"no", u"off", u"0", u"0.0"]:
				truePath = fullPathNoExt + u"+true" + fileExt
				return truePath
#
# Although I like the idea of this additional check (before returning onPath),
# mirroring the logic in Indigo Touch isn't possible since it cannot check
# to see if the file exists before requesting it (and performing multiple
# HTTP requests is not good for performance). So for consistency, we do not
# do this check here either; meaning that device images should always have
# a +on variant if they use the simple heuristic.
#
#				USER_STATIC_FOLDER_PATH = cherrypy.config.get("user_static_folder_path")
#				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, truePath)):
#					return truePath
		return fullPath
	elif controlType == 100:	# It's a static image/control
		fullPath += u"static/" + filename
		return fullPath

	return "ERROR: unknown control type"

####################
# This method will return the path to the closest matching image file
# based on the control type and value. There are several heuristics in here
# so that choosing the best image is a bit easier.
####################
def GetImagePathWithResolver(controlType, name, value, filename):
	fullPath = u"images/controls/"
	# DebugLog("\n\nresolving type: %s, name: %s, value: %s, filename: %s\n\n" % (str(controlType), str(name), str(value), str(filename)))

	# We do this here because the config isn't set apparently when this file gets loaded. We only need it in this
	# function anyway.
	USER_STATIC_FOLDER_PATH = cherrypy.config.get("user_static_folder_path")

	# Identify the control type
	if controlType == 1:		# It's a device
		fullPath += u"devices/" + filename
		fullPathNoExt = fullPath[:-4]
		if fullPathNoExt[-1:] == u"+":
			fullPathNoExt = fullPathNoExt[:-1]
		fileExt = fullPath[-4:]
		if name and value:
			# check to see if there's an exact match to the state of the device
			stepPath = fullPathNoExt + u"+" + value + fileExt

			if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):
				return stepPath

			# If the value is an integer, then try:
			#	a) if the value is 0, assume that the default image (with no +something) is right
			#	b) the next highest value in increments of 5
			#	c) the next highest value in increments of 10
			#	d) the next highest value in increments of 20
			#	e) the next highest value in increments of 25
			if isIntCompat(value):
				intValue = int(value)
				if intValue == 0:				# case a)
					return fullPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 5)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case b)
					return stepPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 10)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case c)
					return stepPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 20)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case d)
					return stepPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 25)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case e)
					return stepPath

			# If it's not an Integer or we didn't find a match, then show the +on image unless it is an OFF variant.
			# This logic should mirror Indigo Touch's GraphicController.m _addViewsForPageElem() function.
			# 2011.09.13: IS5.0.0b9/IT1.8 Added additional boolean types to test: no, false, 0.0.
			if value.lower() not in [u"", u"false", u"no", u"off", u"0", u"0.0", u"paused"]:
				onPath = fullPathNoExt + u"+on" + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, onPath)):
					return onPath
		return fullPath
	elif controlType == 2:		# It's a variable
		fullPath += u"variables/" + filename
		fullPathNoExt = fullPath[:-4]
		if fullPathNoExt[-1:] == u"+":
			fullPathNoExt = fullPathNoExt[:-1]
		fileExt = fullPath[-4:]
		if name and value:
			stepPath = fullPathNoExt + u"+" + value + fileExt
			if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):
				return stepPath

			# If the value is an integer, then try:
			#	a) if the value is 0, assume that the default image (with no +something) is right
			#	b) the next highest value in increments of 5
			#	c) the next highest value in increments of 10
			#	d) the next highest value in increments of 20
			#	e) the next highest value in increments of 25
			if isIntCompat(value):
				intValue = int(value)
				if intValue == 0:				# case a)
					return fullPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 5)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case b)
					return stepPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 10)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case c)
					return stepPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 20)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case d)
					return stepPath
				stepPath = fullPathNoExt + u"+" + str(GetNextIntegerStep(intValue, 25)) + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, stepPath)):	# case e)
					return stepPath

			# If it's not an Integer or we didn't find a match, then show the +true image unless it is a FALSE variant.
			# This logic should mirror Indigo Touch's GraphicController.m _addViewsForPageElem() function.
			if value.lower() not in [u"", u"false", u"no", u"off", u"0", u"0.0"]:
				truePath = fullPathNoExt + u"+true" + fileExt
				if os.path.exists(os.path.join(USER_STATIC_FOLDER_PATH, truePath)):
					return truePath
		return fullPath
	elif controlType == 100:	# It's a static image/control
		fullPath += u"static/" + filename
		return fullPath

	return "ERROR: unknown control type"


