####################
# Copyright (c) 2021, Perceptive Automation, LLC. All rights reserved.
# http://www.indigodomo.com
#
# Redistribution of this source file, its binary forms, and images are not allowed.
# We are working on our open source license terms, but at this point the source files,
# its binary forms, and all images cannot be redistributed without written permission
# from both Perceptive Automation, LLC. and Andy Turner, HighEarthOrbit software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
####################

####################
## IMPORTS
import cherrypy
from cherrypy import _cperror

import indigoconn as ic
import indigodb as idb
import basereqhandler as brh

import os
import sys
import imp
import fnmatch

####################
## CONSTANTS

####################
## GLOBALS

####################
## EXCEPTIONS
class PluginError(Exception):
	def __init__(self, value):
		self.value = value	# unicode string

####################
class PluginInfo:
	""" Document me. """

	def __init__(self, name, description, pluginWebPath, showOnMainPage):
		if not description:
			description = u"&nbsp;"
		elif not name:
			name = u"unknown plugin"
		self.name = name
		self.description = description
		self.pluginWebPath = pluginWebPath
		self.showOnMainPage = showOnMainPage

####################
class NullRoot:
	pass

def null_logger(message):
	pass

####################
class LegacyWebPluginMgr:
	""" Document me. """
	
	def _IsTrueSubclass(self, class1, class2):
		# Unlike issubclass, this returns False if the classes are equal.
		return class1 != class2 and issubclass(class1, class2) and callable(class1)

	def _IterateDirectories(self, dir):
		for dirpath, dirnames, filenames in os.walk(dir):
			for dirname in dirnames:
				folderpath = os.path.join(dirpath,dirname)
				yield folderpath
			break	# just get 1st level of folders -- caller handles deeper folders
	
	def __init__(self, logger):
		self._exposedFolders = []
		self._pluginList = []
		self._infoList = []
		self.logger = logger

	def LoadWebPlugins(self, basedir, app_root_path):
		# Import user-defined plugin modules (inside plugins folder).
		try:
			sys.path.insert(0, basedir)
			app_root_path = app_root_path.rstrip('/ ')	# remove trailing slash
			self._exposedFolders = []
			self._pluginList = []
			self._infoList = []
			for folderpath in self._IterateDirectories(basedir):
				for dirpath, dirnames, filenames in os.walk(folderpath):
					if dirpath.find(basedir) != 0:	# sanity check
						self.logger.info(u"module folder path does not contain plugin path")
						continue
					modulename = dirpath[len(basedir):]
					modulenameForCherryPy = "%s/%s" % (app_root_path, modulename)
					modulenameForImport = modulename.replace("/", ".")
					mountedHandlerIntoTree = False	# can only mount 1 handler per folder

					self.logger.debug("")
					self.logger.debug("processing plugin folder " + modulename)

					# First, find all python module files and install into cherrypy the first BaseRequestHandler subclass.
					for filename in filenames:
						if fnmatch.fnmatch(filename, '*.py'):
							filename_base = os.path.splitext(filename)[0]
							if filename_base == "__init__":
								continue
							try:
								self.logger.debug("- filename: " + filename)
	
								# Runtime version of:
								#	from modulenameForImport import filename_base								
								plugin_parent = __import__(modulenameForImport, globals(), locals(), [filename_base])
								plugin_module = getattr(plugin_parent, filename_base)
								self._pluginList.append(plugin_module)

								# Cache off some info about the plugin.
								name = None
								func = getattr(plugin_module, "PluginName", None)
								if func and callable(func):
									name = func()

								description = None
								func = getattr(plugin_module, "PluginDescription", None)
								if func and callable(func):
									description = func()

								showOnMainPage = False
								func = getattr(plugin_module, "ShowOnMainPage", None)
								if func and callable(func):
									showOnMainPage = func()
								if not showOnMainPage:		# legacy version was called ShowOnControlPageList
									func = getattr(plugin_module, "ShowOnControlPageList", None)
									if func and callable(func):
										showOnMainPage = func()

								if name or description:
									self.logger.debug("-   creating plugin \"%s\" at path %s" % (name, modulenameForCherryPy))
									self._infoList.append(PluginInfo(name, description, modulenameForCherryPy, showOnMainPage))
								else:
									pass		# no name or description, presume plugin wants to be hidden

								# Mount the first request handler subclass in this folder, if it exists.
								for subname in dir(plugin_module):
									submodule = getattr(plugin_module, subname)
									try:
										if self._IsTrueSubclass(submodule, brh.BaseRequestHandler):
											if not mountedHandlerIntoTree:
												self.logger.debug("-   mounting handler class %s at path %s" % (subname, modulenameForCherryPy))
												try:
													cherrypy.tree.mount(submodule(self.logger, null_logger), modulenameForCherryPy)
													mountedHandlerIntoTree = True
												except TypeError:	# the old constructor didn't have the debugLogFunc argument:
													cherrypy.tree.mount(submodule(self.logger), modulenameForCherryPy)
													mountedHandlerIntoTree = True
											else:
												self.logger.info(u"-   failed to mount plugin handler class %s at path %s; only one handler class per folder is allowed" % (subname, modulenameForCherryPy))
										else:
											self.logger.debug("-   issubclass(%s, %s): %s" % (str(submodule), str(brh.BaseRequestHandler), issubclass(submodule, brh.BaseRequestHandler)))
									except TypeError:
										pass	# ignore and go on to the next plugin
							except:
								tb = _cperror.format_exc()
								self.logger.info(tb)

					# Next, search for folders with these reserved names:
					#		'css','js', 'images', 'video', 'static'
					# and enable the static access for them.
					for dirname in dirnames:
						try:
							if dirname in ('css', 'js', 'images', 'video', 'static'):
								full_url = modulenameForCherryPy + "/" + dirname
								
								# Must lookup which mounted application contains this path. It might be the
								# root application, or a plugin application just mounted above.
								target_app_url = cherrypy.tree.script_name(full_url)

								# Now compute the relative path for the static folder based on the target
								# application's path.
								if full_url.find(target_app_url) != 0:	# sanity check
									self.logger.info(u"full static folder path does not contain target app path")
									continue
								relative_url = full_url[len(target_app_url):]

								target_app_name = target_app_url
								if target_app_name is None or target_app_name == "":
									target_app_name = "_root_"
								self.logger.debug("-   enabling static file access to path %s on mounted app %s" % (relative_url, target_app_name))

								enable_staticdir_config = {
									relative_url: {
										'tools.staticdir.on': True,
										'tools.staticdir.root': dirpath,
										'tools.staticdir.dir': dirname
									}
								}
								target_app = cherrypy.tree.apps[target_app_url]
								target_app.merge(enable_staticdir_config)

								self._exposedFolders.append(full_url)
						except:
							tb = _cperror.format_exc()
							self.logger.info(tb)
		except:
			tb = _cperror.format_exc()
			self.logger.info(tb)

	def GetWebPluginInfoList(self):
		return self._infoList

	def IndigoConnected(self):
		for plugin in self._pluginList:
			try:
				func = getattr(plugin, "IndigoConnected", None)
				if func and callable(func):
					func()
				func = getattr(plugin, "PluginName", None)
				if func and callable(func):
					pluginName = func()
					self.logger.info("loaded web plugin \"" + pluginName + "\"")
			except:
				tb = _cperror.format_exc()
				self.logger.info(tb)

		for exposedFolder in self._exposedFolders:
			self.logger.info("enabled browser access to plugin path \"plugins" + exposedFolder + "\"")

	def IndigoDisconnected(self):
		for plugin in self._pluginList:
			try:
				func = getattr(plugin, "IndigoDisconnected", None)
				if func and callable(func):
					func()
			except:
				tb = _cperror.format_exc()
				self.logger.info(tb)
